#drinc:util.g
#drinc:crt.g
#ded.g

/*
 * DED.DRC - main program, status line, top level routines, and selected
 *	     utility routines for the Draco screen editor.
 */

char COPYRIGHT = '\(0xa9)';

extern
    command(ushort cmd)bool,
    editNew(*char fileName)bool,
    freeText()void,
    initLines()void,
    buildStart()void,
    buildLine(*char testPtr)bool,
    buildEnd(**LINELIST pHead, pTail)ulong,
    clearMemory(*LINELIST l)void,
    freeString(*char p)void,
    fixPosition()void,
    forwardOne()void,
    backUpOne()void,
    isBlank()bool,
    isLetDig()bool,
    displayChar(char ch)void,
    displayChars(*char p)void,
    displayString(*char p)void,
    addText(*char txtPtr; bool wantRedraw)void,
    insertChunk(*char txtPtr; bool wantRedraw)void,
    insertBlanks(ushort count; bool wantRedraw)void,
    insertLine()void,
    cursorDown()void,
    displayWindow()void;

ulong OldCursorLine;
ushort OldCursorColumn, HighLightCount;
bool GotError;

/***************************************************************************/

/*
 * COMMANDTABLE SETUP
 */

/*
 * initCommands - set up the command table. This routine is the first one in
 *		  this file, which in turn is the first file given when Ded
 *		  is linked. This insures that this routine is the first
 *		  routine in the object code and thus that it is easy for the
 *		  configuration program to find and modify it.
 */

proc initCommands()void:

    CommandTable := pretend(
	"\(2)"				/* 2 second escape characters */
	"\(0x1b)"			/* c_escapeA = ESCAPE */
	"\(0x18)"			/* c_escapeB = ^X     */
	"\(0x9b)"			/* c_escapeC = <CSI>  */
	"~"				/* c_terminator = ~   */
	"1 \e\e\e\e"			/* second esc chars are '1', ' ' */

    /* First byte in each command entry is a mode byte saying how this
       command is activated. The second byte is the key character associated
       with it. The modes are encoded as:

	00	=> char is direct control or alternate character
	10	=> escapeA followed by the char
	11 - 16 => escapeA, one 2nd escape, the char
	20	=> escapeB followed by the char
	21 - 26 => escapeB, one 2nd escape, the char
	30	=> escapeC followed by the char
	31 - 36 => escapeC, one 2nd escape, the char
	64 - 100 => must have 'c_terminator' after the sequence
     */

	    "\(30)D"		/* LEFT */
	    "\(30)C"		/* RIGHT */
	    "\(30)A"		/* UP */
	    "\(30)B"		/* DOWN */
	    "\(00)\(0x17)"	/* SKIPWORD */
	    "\(00)\(0x02)"	/* BACKWORD */
	    "\(00)\(0x01)"	/* BOL */
	    "\(00)\(0x05)"	/* EOL */
	    "\(00)\(0x0c)"	/* NEXTLINE */
	    "\(00)\(0x14)"	/* TOP */
	    "\(00)\(0x1a)"	/* BOTTOM */
	    "\(30)T"		/* PAGEUP */
	    "\(30)S"		/* PAGEDOWN */
	    "\(32)A"		/* HOME */
	    "\(32)@"		/* GOTO */
	    "\(00)\(0x07)"	/* GOPOINT */
	    "\(00)\(0x7f)"	/* DELCHAR */
	    "\(00)\b"		/* BACKSPACE */
	    "\(00)\t"		/* TAB */
	    "\(00)\(0x04)"	/* DELWORD */
	    "\(00)\(0x0b)"	/* DELBACKWORD */
	    "\(10)e"		/* DELEOL */
	    "\(00)\(0x1c)"	/* INSERTSPECIAL */
	    "\(10)u"		/* UNDO */
	    "\(10)r"		/* REDRAW */
	    "\(10)l"		/* DELLINE */
	    "\(00)\r"		/* RETURN */
	    "\(10)\(0x0d)"	/* INSERTLINE */
	    "\(00)\(0x0a)"	/* JOIN */
	    "\(00)\(0x13)"	/* SPLIT */
	    "\(00)\(0x0e)"	/* INDENT */
	    "\(00)\(0x0f)"	/* OUTDENT */
	    "\(10) "		/* ADJUSTLINE */
	    "\(10)c"		/* GLOBALCHANGE */
	    "\(00)\(0x11)"	/* QUERYREPLACE */
	    "\(00)\(0x06)"	/* FIND */
	    "\(00)\(0x12)"	/* REVERSEFIND */
	    "\(94)?"		/* POINT */
	    "\(10)\t"		/* ADJUST */
	    "\(10)<"		/* SHIFTLEFT */
	    "\(10)>"		/* SHIFTRIGHT */
	    "\(00)\(0x19)"	/* YANK */
	    "\(10)C"		/* COPY */
	    "\(00)\(0x10)"	/* PUT */
	    "\(10)f"		/* FORMAT */
	    "\(10)P"		/* PRINT */
	    "\(10)W"		/* WRITE */
	    "\(10)R"		/* READ */
	    "\(10)D"		/* DELETEFILE */
	    "\(00)\(0x1e)"	/* SCROLLUP */
	    "\(00)\(0x16)"	/* SCROLLDOWN */
	    "\(10)x"		/* EXIT */
	    "\(10)q"		/* QUIT */
	    "\(10)s"		/* SET */
	    "\(10)b"		/* BLANKMODE */
	    "\(10)A"		/* CLEARALL */
	    "\(10)S"		/* TABSET */
	    "\(10)T"		/* TABCLEAR */
	    "\(94)0"		/* INSERT */
	    "\(95)0"		/* INSERT */
	    "\(94)1"		/* WRAP */
	    "\(94)2"		/* MULTI */
	    "\(20)\(0x16)"	/* VISITFILE */
	    "\(20)D"		/* DELETEBUFFER */
	    "\(20)\(0x0f)"	/* OTHERBUFFER */
	    "\(20)d"		/* DELETEWINDOW */
	    "\(20)t"		/* SPLITWINDOW */
	    "\(20)s"		/* ONEWINDOW */
	    "\(20)n"		/* NEXTWINDOW */
	    "\(20)p"		/* PREVIOUSWINDOW */
	    "\(20)\(0x0d)"	/* UPDATEFILES */
	    "\(20)\(0x02)"	/* BUFFERLIST */
	    "\(00)\(0x15)"	/* PREVLINE */
	    "\(10)B"		/* BATCHCHANGE */
	    "\(10)X"		/* NEWSHELL */
	, *COMMANDTABLE
    );
corp;

/***************************************************************************/

/*
 * ERROR MESSAGE ROUTINES.
 */

/*
 * noSpace - an operation has run out of memory space.
 */

proc noSpace()void:

    NoMemory := true;
corp;

/*
 * tooLong - an operation resulted in a line that's too long.
 */

proc tooLong()void:

    TooLong := true;
corp;

/*
 * clearError - clear out the message area.
 */

proc clearError()void:

    CRT_ClearLine(NLines - 1);
    GotError := false;
corp;

/*
 * putError - display an error message in the message line.
 */

proc putError(*char message)void:

    clearError();
    CRT_PutChars(message);
    GotError := true;
corp;
	
/*
 * noFile - an operation needing a file, when no file is active.
 */

proc noFile()void:

    putError("No file to edit yet!");
corp;

/*
 * noRegion - complain that no first end has been selected.
 */

proc noRegion()void:

    putError("No first region end yet.");
corp;

/*
 * noText - complain that there is no text in the region.
 */

proc noText()void:

    putError("Selected region is empty.");
corp;

/*
 * putNoMemory - put out the message.
 */

proc putNoMemory()void:

    putError("Out of memory.\(BEEP)");
corp;

/*
 * update - put the cursor in the proper place.
 */

proc update()void:

    CRT_Move(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,
	     CurrentBuffer.b_cursorColumn);
corp;

/***************************************************************************/

/*
 * ROUTINES FOR DEALING WITH THE STATUS LINE.
 */

/*
 * highLight - enter highlight mode if not already in it.
 */

proc highLight()void:

    if HighLightCount = 0 then
	CRT_EnterHighLight();
    fi;
    HighLightCount := HighLightCount + 1;
corp;

/*
 * lowLight - exit highlight mode if in it.
 */

proc lowLight()void:

    HighLightCount := HighLightCount - 1;
    if HighLightCount = 0 then
	CRT_ExitHighLight();
    fi;
corp;

/*
 * putNumber - write a positive integer to the screen.
 */

proc putNumber(register ulong n; register ushort digits)void:
    [11] char buffer;
    register *char p;

    p := &buffer[5];
    p* := '\e';
    while digits ~= 0 do
	digits := digits - 1;
	p := p - 1;
	p* := n % 10 + '0';
	n := n / 10;
    od;
    CRT_PutChars(p);
corp;

/*
 * colStatus - fix the column display.
 */

proc colStatus()void:

    highLight();
    CRT_Move(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowHeight, COLCOL);
    putNumber(CurrentBuffer.b_cursorColumn + 1, 2);
    lowLight();
corp;

/*
 * lineStatus - fix the line number display.
 */

proc lineStatus()void:

    highLight();
    CRT_Move(CurrentBuffer.b_windowPos+CurrentBuffer.b_windowHeight, LINECOL);
    putNumber(CurrentBuffer.b_cursorLine, 5);
    lowLight();
corp;

/*
 * displayFlag - display a status flag.
 */

proc displayFlag(ushort column; bool flag; char set)void:

    highLight();
    CRT_Move(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowHeight, column);
    CRT_PutChar(if flag then set else '-' fi);
    lowLight();
corp;

/*
 * insertStatus - display the insert mode flag.
 */

proc insertStatus()void:

    displayFlag(INSERTCOL, CurrentBuffer.b_insertMode, 'I');
corp;

/*
 * wrapStatus - display the wrap mode flag.
 */

proc wrapStatus()void:

    displayFlag(WRAPCOL, CurrentBuffer.b_wrapMode, 'W');
corp;

/*
 * multiStatus - display the multcolumn flag.
 */

proc multiStatus()void:

    displayFlag(MULTICOL, CurrentBuffer.b_multiColumn, 'M');
corp;

/*
 * regionStatus - display the status of the region end.
 */

proc regionStatus()void:

    displayFlag(REGIONCOL, CurrentBuffer.b_gotRegion, '1');
corp;

/*
 * fullStatus - display the full status line.
 */

proc fullStatus()void:
    register ushort i;

    CRT_ClearLine(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowHeight);
    highLight();
    CRT_PutChar('-');
    insertStatus();
    wrapStatus();
    multiStatus();
    regionStatus();
    CRT_PutChar('-');
    lineStatus();
    CRT_PutChar('-');
    colStatus();
    CRT_PutChar('-');
    CRT_PutChars(CurrentBuffer.b_fileName);
    i := CharsLen(CurrentBuffer.b_fileName) + NAMECOL;
    while i ~= NCOLUMNS do
	i := i + 1;
	CRT_PutChar('-');
    od;
    lowLight();
corp;

/***************************************************************************/

/*
 * ROUTINES FOR GOING TO AND FROM THE INTERNAL BLANK-COMPRESSED FORMAT.
 */

/*
 * compress - put appropriate compressed blanks into the line in LineBuffer.
 *	      A fair attempt is made to handle quoted strings.
 *	      Return length + 1 of resulting string.
 */

proc compress()ushort:
    char NOQUOTE = '\(0x00)';
    register *char p1, p2;
    register ushort column;
    ushort blankCount, newCount;
    register char ch, quote;

    p1 := &LineBuffer[0];
    p2 := p1;
    column := 0;
    blankCount := 0;
    quote := NOQUOTE;
    while
	ch := p2*;
	p2 := p2 + 1;
	if ch = quote then
	    quote := NOQUOTE;
	elif ch = '"' and quote = NOQUOTE then
	    quote := ch;
	fi;
	newCount :=
	    if quote ~= NOQUOTE then
		0
	    elif ch = ' ' then
		1
	    elif ch >= '\(0x81)' then
		ch - '\(0x80)'
	    elif ch = '\t' then
		((column + 8) & 0xf8) - column
	    else
		0
	    fi;
	if newCount = 0 then
	    if column < NCOLUMNS and blankCount ~= 0 then
		p1* := blankCount + '\(0x80)';
		p1 := p1 + 1;
	    fi;
	    blankCount := 0;
	    if column < NCOLUMNS then
		p1* := ch;
		p1 := p1 + 1;
	    else
		p1* := '\e';
		tooLong();
	    fi;
	    column := column + 1;
	else
	    blankCount := newCount + blankCount;
	    column := newCount + column;
	    if column >= NCOLUMNS then
		column := NCOLUMNS - 1;
	    fi;
	fi;
	ch ~= '\e'
    do
    od;
    p1 - &LineBuffer[0]
corp;

/*
 * expand - expand a given string into LineBuffer.
 *	    Return the resulting length.
 */

proc expand(register *char p)ushort:
    char NOQUOTE = '\(0x00)';
    register *char q;
    register char ch, quote;

    q := &LineBuffer[0];
    quote := NOQUOTE;
    while
	ch := p*;
	p := p + 1;
	if ch = quote then
	    quote := NOQUOTE;
	elif ch = '"' and quote = NOQUOTE then
	    quote := ch;
	fi;
	if quote = NOQUOTE and ch >= '\(0x80)' then
	    while ch > '\(0x80)' do
		ch := ch - 1;
		q* := ' ';
		q := q + 1;
	    od;
	else
	    q* := ch;
	    q := q + 1;
	fi;
	ch ~= '\e'
    do
    od;
    q - &LineBuffer[0] - 1
corp;

/*
 * unFixLine - move the current line to LineBuffer.
 */

proc unFixLine()void:

    if LineFixed then
	LineFixed := false;
	LineChanged := false;
	LineLength := expand(
	    if CurrentBuffer.b_currentLine = nil then
		""
	    else
		CurrentBuffer.b_currentLine*.l_data
	    fi);
	CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
    fi;
corp;

/*
 * fixLine - move the line in LineBuffer back into the list of lines.
 */

proc fixLine()void:
    register *char p;

    if not LineFixed then
	LineFixed := true;
	if LineChanged then
	    CurrentBuffer.b_changeCount := CurrentBuffer.b_changeCount + 1;
	    initLines();
	    LineLength := compress();
	    if not TooLong then
		p := pretend(Malloc(LineLength), *char);
		if p = nil then
		    noSpace();
		    TooLong := true;
		fi;
	    fi;
	    if TooLong then
		/* bad line, have to undo changes to this line. */
		CRT_ClearLine(CurrentBuffer.b_windowPos +
			      CurrentBuffer.b_windowLine);
		displayString(CurrentBuffer.b_currentLine*.l_data);
	    else
		CharsCopy(p, &LineBuffer[0]);
		freeString(CurrentBuffer.b_currentLine*.l_data);
		CurrentBuffer.b_currentLine*.l_data := p;
	    fi;
	fi;
    fi;
corp;

/***************************************************************************/

/*
 * WORDWRAP/MULTI & ROUTINE USED WHEN A NON-FUNCTION CHARACTER IS TYPED.
 */

/*
 * wrapMulti - handle wordwrap and multi-column stuff.
 */

proc wrapMulti(register char ch)char:
    register *char p, q;
    register char saveCh;

    p := CursorPos;
    saveCh := CursorPos*;
    if CurrentBuffer.b_multiColumn then
	CursorPos* := '\e';
    fi;
    if ch ~= ' ' and ch ~= '\r' then
	backUpOne();
	while CurrentBuffer.b_cursorColumn ~= CurrentBuffer.b_indentAmount
		and isLetDig() do
	    backUpOne();
	od;
	if CurrentBuffer.b_cursorColumn = CurrentBuffer.b_indentAmount then
	    CursorPos := p;
	    CurrentBuffer.b_cursorColumn := p - &LineBuffer[0];
	else
	    forwardOne();
	fi;
    fi;
    q := pretend(Malloc(CharsLen(CursorPos) + 1), *char);
    if q = nil then
	noSpace();
	p* := saveCh;
    else
	LineChanged := true;
	CharsCopy(q, CursorPos);
	if CurrentBuffer.b_cursorColumn ~= CurrentBuffer.b_indentAmount and
		not CurrentBuffer.b_multiColumn then
	    backUpOne();
	    while CurrentBuffer.b_cursorColumn ~= CurrentBuffer.b_indentAmount
		    and isBlank() do
		backUpOne();
	    od;
	    if not isBlank() then
		forwardOne();
	    fi;
	fi;
	update();
	if CurrentBuffer.b_multiColumn and saveCh ~= '\e' then
	    while CursorPos* ~= '\e' do
		CursorPos* := ' ';
		CRT_PutChar(' ');
		forwardOne();
	    od;
	else
	    CRT_ClearTail();
	    CursorPos* := '\e';
	fi;
	p* := saveCh;
	if CurrentBuffer.b_cursorLine >= CurrentBuffer.b_lineCount or
		not CurrentBuffer.b_multiColumn then
	    insertLine();
	    update();
	    insertChunk(q, true);
	else
	    cursorDown();
	    unFixLine();
	    if LineLength < CurrentBuffer.b_indentAmount then
		CurrentBuffer.b_cursorColumn := LineLength;
		CursorPos := &LineBuffer[LineLength];
		update();
		insertBlanks(CurrentBuffer.b_indentAmount - LineLength, true);
	    else
		CurrentBuffer.b_cursorColumn := CurrentBuffer.b_indentAmount;
		CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
		update();
	    fi;
	    addText(q, true);
	    if LineLength < CurrentBuffer.b_cursorColumn then
		LineLength := CurrentBuffer.b_cursorColumn;
		CursorPos* := '\e';
	    fi;
	fi;
	if q* = '\e' and ch = ' ' then
	    /* don't put a blank on a new "empty" column */
	    ch := '\e';
	fi;
	freeString(q);
    fi;
    ch
corp;

/*
 * dataChar - a regular data character to put into the file.
 */

proc dataChar(register char ch)void:
    register *char p;

    initLines();
    unFixLine();
    fixPosition();
    if CurrentBuffer.b_cursorColumn = CurrentBuffer.b_wrapColumn and
	    CurrentBuffer.b_wrapMode then
	ch := wrapMulti(ch);
    fi;
    if ch = '\e' then
	/* don't put a blank at the start of a new line from wrapping */
    elif
	if CurrentBuffer.b_insertMode then
	    LineLength
	else
	    CurrentBuffer.b_cursorColumn
	fi ~= NCOLUMNS - 1
    then
	p := &LineBuffer[LineLength];
	(p + 1)* := '\e';
	LineChanged := true;
	update();
	if CurrentBuffer.b_insertMode then
	    BlockCopyB(pretend(p, *byte), pretend(p - 1, *byte),p - CursorPos);
	    CRT_InsertChars(1);
	    update();
	fi;
	CursorPos* := ch;
	displayChar(ch);
	forwardOne();
	if CurrentBuffer.b_cursorColumn > LineLength or
		CurrentBuffer.b_insertMode then
	    LineLength := LineLength + 1;
	fi;
    fi;
corp;

/***************************************************************************/

/*
 * FILE INPUT/OUTPUT ROUTINES.
 */

/*
 * readData - read some data from a file.
 */

proc readData(**LINELIST head, tail; ulong maxCount)ulong:
    register ulong possibleLength;

    putError("Reading file...");
    CRT_Flush();
    buildStart();
    while
	if maxCount = 0 then
	    false
	else
	    possibleLength := LineRead(Chin, &LineBuffer[0], NCOLUMNS - 1);
	    if possibleLength = LINE_EOF then
		false
	    else
		if possibleLength > NCOLUMNS - 1 then
		    tooLong();
		    possibleLength := NCOLUMNS - 1;
		fi;
		LineBuffer[possibleLength] := '\e';
		pretend(compress(), void);
		maxCount := maxCount - 1;
		buildLine(&LineBuffer[0])
	    fi
	fi
    do
    od;
    if not NoMemory and not TooLong then
	clearError();
    fi;
    buildEnd(head, tail)
corp;

/*
 * readFile - read in a file, assumed to be opened on the input channel.
 */

proc readFile()void:

    CurrentBuffer.b_lineCount :=
	readData(&CurrentBuffer.b_lines, &CurrentBuffer.b_regionPtr, 0xffff);
corp;

/*
 * writeFile - write the edited file out to the current output file.
 */

proc writeFile(register *LINELIST l)bool:
    char NOQUOTE = '\(0x00)';
    register *char p, q;
    register ushort column, length, blankCount;
    register char ch;
    char quote;
    bool diskFull;

    putError("Writing file ");
    CRT_PutChars(CurrentBuffer.b_fileName);
    CRT_PutChars("...");
    CRT_Flush();
    diskFull := false;
    while not diskFull and l ~= nil do
	p := l*.l_data;
	length := CharsLen(l*.l_data);
	if CurrentBuffer.b_tabMode ~= COUNTS then
	    length := expand(p);
	    p := &LineBuffer[0];
	    if CurrentBuffer.b_tabMode = TABS then
		q := p;
		column := 0;
		blankCount := 0;
		length := 0;
		quote := NOQUOTE;
		while
		    ch := p*;
		    p := p + 1;
		    column := column + 1;
		    ch ~= '\e'
		do
		    if ch = quote then
			quote := NOQUOTE;
		    elif ch = '"' and quote = NOQUOTE then
			quote := ch;
		    fi;
		    if ch ~= ' ' or quote ~= NOQUOTE then
			while blankCount ~= 0 do
			    blankCount := blankCount - 1;
			    q* := ' ';
			    q := q + 1;
			    length := length + 1;
			od;
			q* := ch;
			q := q + 1;
			length := length + 1;
		    else
			blankCount := blankCount + 1;
			if column & 0x7 = 0 then
			    q* := if blankCount = 1 then ' ' else '\t' fi;
			    q := q + 1;
			    length := length + 1;
			    blankCount := 0;
			fi;
		    fi;
		od;
		p := &LineBuffer[0];
	    fi;
	fi;
	diskFull := LineWrite(Chout, p, length) ~= length;
	l := l*.l_next;
    od;
    if not close(Chout) then
	diskFull := true;
    fi;
    clearError();
    if diskFull then
	putError("Disk full.\(BEEP)");
	false
    else
	true
    fi
corp;

/*
 * VARIOUS ROUTINES FOR GETTING INPUT FROM THE USER.
 */

/*
 * getString - get a string in the error input area.
 */

proc getString(*char prompt)void:
    register ushort column;
    register char ch;

    fixLine();
    putError(prompt);
    column := CharsLen(prompt);
    CursorPos := &LineBuffer[0];
    while
	ch := CRT_GetChar();
	ch ~= '\r'
    do
	if ch = '\b' then
	    if CursorPos ~= &LineBuffer[0] then
		CursorPos := CursorPos - 1;
		column := column - 1;
		CRT_PutChars("\b \b");
	    fi;
	elif ch >= ' ' and ch <= '~' and column ~= NCOLUMNS - 1 then
	    column := column + 1;
	    CursorPos* := ch;
	    CursorPos := CursorPos + 1;
	    CRT_PutChar(ch);
	fi;
    od;
    CursorPos* := '\e';
    clearError();
corp;

/*
 * getFileName - request a file name from the user.
 */

proc getFileName(*char prompt)bool:
    register *char p, q;

    getString(prompt);
    p := &LineBuffer[0];
    while p* = ' ' do
	p := p + 1;
    od;
    if p* = '\e' then
	false
    else
	q := &LineBuffer[0];
	while p* ~= '\e' do
	    q* := p*;
	    q := q + 1;
	    p := p + 1;
	od;
	while
	    q := q - 1;
	    q* = ' '
	do
	od;
	(q + 1)* := '\e';
	true
    fi
corp;

/*
 * getNumber1 - lower level of getNumber and getParameter.
 */

proc getNumber1()ulong:
    register ulong n;
    register ushort cnt;
    register char ch;

    n := 0;
    cnt := 0;
    while
	ch := CRT_GetChar();
	ch ~= '\r'
    do
	if ch = '\b' and cnt ~= 0 then
	    cnt := cnt - 1;
	    n := n / 10;
	    CRT_PutChars("\b \b");
	elif ch >= '0' and ch <= '9' and
		(n < 429496729 or n = 429496729 and ch < '6') and
		cnt ~= 20 then
	    cnt := cnt + 1;
	    n := n * 10 + (ch - '0');
	    CRT_PutChar(ch);
	fi;
    od;
    clearError();
    n
corp;

/*
 * getNumber - get a number in the error input area.
 */

proc getNumber(*char prompt)ulong:

    putError(prompt);
    getNumber1()
corp;

/*
 * setParameter - set a numeric parameter.
 */

proc setParameter(*char prompt; *ushort var; ushort min, max)void:
    ushort n;

    clearError();
    CRT_PutChars("New ");
    CRT_PutChars(prompt);
    CRT_PutChars(": ");
    n := getNumber1();
    if n < min then
	putError("Value too small.");
    elif n > max then
	putError("Value too large.");
    else
	var* := n;
    fi;
corp;

/*
 * ask - ask a yes/no question of the user.
 */

proc ask(*char question)bool:
    char ch;

    putError(question);
    ch := CRT_GetChar();
    clearError();
    ch = 'y' or ch = 'Y'
corp;

/*
 * checkAbandon - confirm that changes are to be abandoned.
 */

proc checkAbandon()bool:

    fixLine();
    BufferPtr* := CurrentBuffer;
    CurrentBuffer.b_changeCount = 0 or ask("Abandon changes (Y/N)? ")
corp;

/***************************************************************************/

/*
 * THE TOP LEVEL ROUTINES FOR THE EDITOR - MAIN FILE AND COMMAND HANDLING.
 */

/*
 * newBuffer - this buffer has just been made the current one.
 */

proc newBuffer()void:

    LRUCount := LRUCount + 1;
    CurrentBuffer.b_LRU := LRUCount;
    OldCursorColumn := CurrentBuffer.b_cursorColumn;
    OldCursorLine := CurrentBuffer.b_cursorLine;
corp;

/*
 * getCommandChar - get a character for a command.
 */

proc getCommandChar()char:
    char ch;

    if NoMemory then
	NoMemory := false;
	putNoMemory();
    else
	if TooLong then
	    TooLong := false;
	    putError("Line too long.\(BEEP)");
	fi;
    fi;
    if BufferList ~= nil then
	if CurrentBuffer.b_cursorColumn ~= OldCursorColumn then
	    OldCursorColumn := CurrentBuffer.b_cursorColumn;
	    colStatus();
	fi;
	if CurrentBuffer.b_cursorLine ~= OldCursorLine then
	    OldCursorLine := CurrentBuffer.b_cursorLine;
	    lineStatus();
	fi;
	update();
    fi;
    ch := CRT_GetChar();
    if GotError then
	clearError();
	update();
	CRT_Flush();
    fi;
    ch
corp;

/*
 * editFile - get ready to edit the file whose name is pointed to by
 *	      the passed parameter. Return 'false' if we have a problem
 */

proc editFile(*char name)bool:

    CurrentBuffer.b_cursorColumn := 0;
    CurrentBuffer.b_windowLine := 0;
    CurrentBuffer.b_cursorLine := 1;
    CurrentBuffer.b_changeCount := 0;
    CurrentBuffer.b_gotRegion := false;
    CurrentBuffer.b_indentAmount := 0;
    CurrentBuffer.b_insertMode := false;
    CurrentBuffer.b_wrapMode := false;
    CurrentBuffer.b_multiColumn := false;
/*
    CurrentBuffer.b_tabMode := TABS;
*/
CurrentBuffer.b_tabMode := BLANKS;
    CurrentBuffer.b_indentIncrement := DEFAULTINDENTINCREMENT;
    CurrentBuffer.b_wrapColumn := DEFAULTWRAPCOLUMN;
    CurrentBuffer.b_lines := nil;
    CurrentBuffer.b_currentLine := nil;
    CurrentBuffer.b_regionPtr := nil;
    CurrentBuffer.b_regionPos := nil;
    CurrentBuffer.b_fileName := pretend(Malloc(CharsLen(name) + 1), *char);
    if CurrentBuffer.b_fileName = nil then
	BufferPtr* := CurrentBuffer;
	putNoMemory();
	false
    else
	CharsCopy(CurrentBuffer.b_fileName, name);
	CRT_ClearLines(CurrentBuffer.b_windowPos,CurrentBuffer.b_windowHeight);
	fullStatus();
	if open(Chin, Fyle, CurrentBuffer.b_fileName) then
	    /* file exists - read it in */
	    CurrentBuffer.b_newFile := false;
	    readFile();
	    close(Chin);
	else
	    /* editing a brand new file */
	    CurrentBuffer.b_newFile := true;
	    CurrentBuffer.b_lines := nil;
	    CurrentBuffer.b_lineCount := 0;
	fi;
	CurrentBuffer.b_currentLine := CurrentBuffer.b_lines;
	displayWindow();
	BufferPtr* := CurrentBuffer;
	true
    fi
corp;

/*
 * commands - loop, getting and processing editing commands.
 */

proc commands()void:
    register *COMMANDENTRY p;
    register ushort baseMode, mode, cmd;
    register char ch;
    bool earlyTerminator, quit;

    quit := false;
    while not quit do
	ch := getCommandChar();
	if ch >= ' ' and ch <= '~' then
	    if BufferList = nil then
		noFile();
	    else
		dataChar(ch);
	    fi;
	else
	    earlyTerminator := false;
	    mode := 0;
	    if ch = CommandTable*.c_escapeA or
		    ch = CommandTable*.c_escapeB or
		    ch = CommandTable*.c_escapeC then
		baseMode :=
		    if ch = CommandTable*.c_escapeA then
			10
		    elif ch = CommandTable*.c_escapeB then
			20
		    else
			30
		    fi;
		ch := CRT_GetChar();
		while mode ~= CommandTable*.c_secondCount and
			ch ~= CommandTable*.c_escape2[mode] do
		    mode := mode + 1;
		od;
		mode :=
		    if mode = CommandTable*.c_secondCount then
			baseMode
		    else
			ch := CRT_GetChar();
			if ch = CommandTable*.c_terminator then
			    earlyTerminator := true;
			    ch := CommandTable*.c_escape2[mode];
			    baseMode
			else
			    baseMode + mode + 1
			fi
		    fi;
	    fi;
	    cmd := 0;
	    p := &CommandTable*.c_commands[0];
	    while cmd ~= COMMANDCOUNT and
		    (p*.c_mode & 0x3f ~= mode or p*.c_char ~= ch) do
		p := p + sizeof(COMMANDENTRY);
		cmd := cmd + 1;
	    od;
	    if cmd = COMMANDCOUNT then
		putError("Invalid command.");
	    else
		if p*.c_mode >= 0x40 then
		    /* flag for terminator character is on */
		    if not earlyTerminator and
			    CRT_GetChar() ~= CommandTable*.c_terminator then
			putError("Invalid command.");
		    else
			if not command(cmd) then
			    quit := true;
			fi;
		    fi;
		else
		    if earlyTerminator then
			putError("Invalid command.");
		    else
			if not command(cmd) then
			    quit := true;
			fi;
		    fi;
		fi;
	    fi;
	fi;
    od;
corp;

/*
 * updateFile - write the updated form of the current file back to disk.
 */

proc updateFile()bool:

    CharsCopy(&TempBuffer[0], CurrentBuffer.b_fileName);
    CharsConcat(&TempBuffer[0], ".BAK");
    /* we don't care if there was an old .BAK or not */
    pretend(FileDestroy(&TempBuffer[0]), void);
    if CurrentBuffer.b_newFile or
	    FileRename(CurrentBuffer.b_fileName, &TempBuffer[0]) then
	if FileCreate(CurrentBuffer.b_fileName) then
	    if open(Chout, Fyle, CurrentBuffer.b_fileName) then
		if writeFile(CurrentBuffer.b_lines) then
		    CurrentBuffer.b_changeCount := 0;
		    BufferPtr* := CurrentBuffer;
		    true
		else
		    pretend(FileDestroy(CurrentBuffer.b_fileName), void);
		    pretend(FileRename(&TempBuffer[0],
				       CurrentBuffer.b_fileName), void);
		    false
		fi
	    else
		putError("Can't open to write");
		pretend(FileDestroy(CurrentBuffer.b_fileName), void);
		pretend(FileRename(&TempBuffer[0],
				   CurrentBuffer.b_fileName), void);
		false
	    fi
	else
	    putError("Can't create to write");
	    pretend(FileRename(&TempBuffer[0],
			       CurrentBuffer.b_fileName), void);
	    false
	fi
    else
	putError("Can't rename to write");
	false
    fi
corp;

/*
 * main - initialize and coordinate everything.
 */

proc main()void:
    register *BUFFERLIST b, temp;
    *char parameter;
    register ushort col;
    bool problem;

    MerrorSet(true);
    CRT_Initialize("Ded V\(VERSION1 + '0').\(VERSION2 + '0') "
		   "\(COPYRIGHT) 1986 by Chris Gray",
		   0, 1);
    NLines := CRT_NLines();
    if CRT_NCols() < NCOLUMNS then
	writeln("Ded: can't open ", NCOLUMNS, " column window - aborting.");
	CRT_Abort();
    fi;
    HighLightCount := 0;
    initCommands();
    LineFixed := true;
    GotText := false;
    TextChars := nil;
    TextLines := nil;
    Searching := false;
    Changing := false;
    NoMemory := false;
    TooLong := false;
    GotError := false;
    for col from 0 upto NCOLUMNS - 1 do
	TabStops[col] := col % 8 = 0;
    od;
    BufferList := nil;
    ScreenList := nil;
    LRUCount := 0;
    problem := false;
    while
	parameter := GetPar();
	if parameter = nil then
	    false
	elif editNew(parameter) then
	    true
	else
	    problem := true;
	    false
	fi
    do
    od;
    if not problem then
	commands();
	freeText();
    fi;
    if GotError then
	update();
	pretend(CRT_ReadChar(), void);
    fi;
    if BufferList ~= nil then
	BufferList*.b_prev*.b_next := nil;
	b := BufferList;
	while b ~= nil do
	    clearMemory(b*.b_lines);
	    freeString(b*.b_fileName);
	    temp := b;
	    b := b*.b_next;
	    free(temp);
	od;
    fi;
    CRT_Terminate();
corp;
