#drinc:util.g
#drinc:crt.g
#ded.g

/*
 * dedb.d - batch change command for Draco screen editor.
 */

type
    ChangeList_t = struct {
	*ChangeList_t cl_next;
	*char cl_from, cl_to;
	uint cl_fromLength, cl_toLength;
    };

*ChangeList_t ChangeList;

extern
    fixLine()void,
    unFixLine()void,
    getFileName(*char prompt)bool,
    putError(*char message)void,
    putNumber(ulong n; ushort digits)void,
    noSpace()void,
    tooLong()void,
    cursorDown()void,
    displayString(*char str)void;

/*
 * doChanges - actually do a batch of changes.
 */

proc doChanges()void:
    *char p;
    register *char q, r;
    register *ChangeList_t cl;
    register ulong tailLength;
    bool hadError, doneChange;

    hadError := false;
    while not hadError do
	unFixLine();
	p := &LineBuffer[0];
	while p* ~= '\e' and not hadError do
	    cl := ChangeList;
	    doneChange := false;
	    while cl ~= nil and not doneChange do
		q := p;
		r := cl*.cl_from;
		while q* = r* and r* ~= '\e' do
		    q := q + 1;
		    r := r + 1;
		od;
		if r* = '\e' then
		    doneChange := true;
		    if LineLength - cl*.cl_fromLength + cl*.cl_toLength >=
			    NCOLUMNS then
			tooLong();
			hadError := true;
		    else
			tailLength := LineLength - (q - &LineBuffer[0]);
			if cl*.cl_toLength < cl*.cl_fromLength then
			    BlockCopy(pretend(p + cl*.cl_toLength, *byte),
				      pretend(p + cl*.cl_fromLength, *byte),
				      tailLength + 1);
			elif cl*.cl_fromLength < cl*.cl_toLength then
			    BlockCopyB(pretend(p+tailLength+cl*.cl_toLength,
					       *byte),
				       pretend(p+tailLength+cl*.cl_fromLength,
					       *byte),
				       tailLength + 1);
			fi;
			LineLength := LineLength - cl*.cl_fromLength +
				cl*.cl_toLength;
			BlockCopy(pretend(p, *byte),
				  pretend(cl*.cl_to, *byte),
				  cl*.cl_toLength);
			LineChanged := true;
		    fi;
		fi;
		cl := cl*.cl_next;
	    od;
	    p := p + 1;
	od;
	if LineChanged then
	    fixLine();
	    CRT_ClearLine(CurrentBuffer.b_windowPos +
			  CurrentBuffer.b_windowLine);
	    displayString(CurrentBuffer.b_currentLine*.l_data);
	    CRT_Flush();
	fi;
	if CurrentBuffer.b_cursorLine < CurrentBuffer.b_lineCount then
	    if not hadError then
		cursorDown();
	    fi;
	else
	    hadError := true;
	fi;
    od;
corp;

/*
 * batchChange - do our thing.
 */

proc batchChange()void:
    register **ChangeList_t pcl;
    register *ChangeList_t cl;
    register ulong length, pos;
    ulong middle;
    register uint lineNumber;
    register char delimiter;
    bool hadError;

    if getFileName("File of substitutions: ") then
	if open(Chin, Fyle, &LineBuffer[0]) then
	    hadError := false;
	    lineNumber := 1;
	    pcl := &ChangeList;
	    ChangeList := nil;
	    while
		if hadError then
		    false
		else
		    length := LineRead(Chin, &LineBuffer[0], NCOLUMNS - 1);
		    length ~= LINE_EOF
		fi
	    do
		if length > NCOLUMNS - 1 then
		    putError("Change line ");
		    putNumber(lineNumber, 4);
		    CRT_PutChars(" too long.");
		    hadError := true;
		elif length < 3 then
		    putError("Change line ");
		    putNumber(lineNumber, 4);
		    CRT_PutChars(" too short.");
		    hadError := true;
		else
		    delimiter := LineBuffer[0];
		    pos := 1;
		    while pos ~= length and LineBuffer[pos] ~= delimiter do
			pos := pos + 1;
		    od;
		    if pos = length then
			putError("Missing second delimiter, line ");
			putNumber(lineNumber, 4);
			hadError := true;
		    else
			LineBuffer[pos] := '\e';
			cl := new(ChangeList_t);
			if cl = nil then
			    noSpace();
			    hadError := true;
			else
			    cl*.cl_from := pretend(Malloc(pos), *char);
			    if cl*.cl_from = nil then
				noSpace();
				free(cl);
				hadError := true;
			    else
				CharsCopy(cl*.cl_from, &LineBuffer[1]);
				cl*.cl_fromLength := pos - 1;
				pos := pos + 1;
				middle := pos;
				while pos ~= length and
					LineBuffer[pos] ~= delimiter do
				    pos := pos + 1;
				od;
				if pos = length then
				    putError("Missing third delimiter, line ");
				    putNumber(lineNumber, 4);
				    Mfree(pretend(cl*.cl_from, *byte),
					  middle - 1);
				    free(cl);
				    hadError := true;
				else
				    cl*.cl_to :=
					pretend(Malloc(pos - middle+1), *char);
				    if cl = nil then
					noSpace();
					Mfree(pretend(cl*.cl_from, *byte),
					      middle - 1);
					free(cl);
					hadError := true;
				    else
					cl*.cl_toLength := pos - middle;
					LineBuffer[pos] := '\e';
					CharsCopy(cl*.cl_to,
						  &LineBuffer[middle]);
					pcl* := cl;
					pcl := &cl*.cl_next;
					cl*.cl_next := nil;
				    fi;
				fi;
			    fi;
			fi;
		    fi;
		fi;
		lineNumber := lineNumber + 1;
	    od;
	    if not hadError then
		doChanges();
	    fi;
	    while ChangeList ~= nil do
		cl := ChangeList;
		ChangeList := cl*.cl_next;
		Mfree(pretend(cl*.cl_to, *byte), CharsLen(cl*.cl_to) + 1);
		Mfree(pretend(cl*.cl_from, *byte), CharsLen(cl*.cl_from) + 1);
		free(cl);
	    od;
	    close(Chin);
	else
	    putError("Can't open file.");
	fi;
    fi;
corp;
