#drinc:util.g
#drinc:crt.g
#ded.g

/*
 * dedc.c - command routines for Draco screen editor.
 */

extern
    DeleteFile(*char fileName)ulong,
    Execute(*char command; ulong inputHandle, outputHandle)bool,

    clearError()void,
    noSpace()void,
    tooLong()void,
    noFile()void,
    insertStatus()void,
    wrapStatus()void,
    multiStatus()void,
    regionStatus()void,
    fullStatus()void,
    update()void,
    putError(*char message)void,
    putNumber(ulong n; ushort digits)void,
    newBuffer()void,
    freeString(*char p)void,
    isBlank()bool,
    isLetDig()bool,
    forwardOne()void,
    backUpOne()void,
    displayChar(char ch)void,
    displayString(*char ptr)void,
    displayWindow()void,
    displayScreen()void,
    displayNew()void,
    displayBottom(ushort start)void,
    initLines()void,
    unFixLine()void,
    fixPosition()void,
    fixLine()void,
    wrapMulti(char ch)char,
    getString(*char prompt)void,
    getFileName(*char prompt)bool,
    getNumber(*char prompt)ulong,
    setParameter(*char prompt; *ushort var; ushort min, max)void,
    ask(*char question)bool,
    updateFile()bool,
    editFile(*char fileName)bool,
    checkAbandon()bool,
    clearMemory(*LINELIST l)void,
    addText(*char textPtr; bool needRedraw)void,
    deleteChunk(*char startPos, endPos; bool needRedraw)void,
    insertChunk(*char chunk; bool needRedraw)void,
    insertBlanks(ushort count; bool needRedraw)void,
    gotoRegion()void,
    grabRegion(bool isYank)void,
    putRegion()void,
    formatRegion()void,
    fileRead()void,
    outputRegion(bool isPrint)void,
    adjustRegion(ushort operationCode)void,
    batchChange()void;

ushort ToLength;		/* length of 'to' string for changes */

/***************************************************************************/

/*
 * CURSOR POSITIONING ROUTINES
 */

/*
 * cursorHome - move window and cursor to first char of first line.
 */

proc cursorHome()void:
    bool needRedraw;

    fixLine();
    needRedraw := CurrentBuffer.b_cursorLine > CurrentBuffer.b_windowHeight or
	    CurrentBuffer.b_cursorLine - 1 ~= CurrentBuffer.b_windowLine;
    CurrentBuffer.b_windowLine := 0;
    CurrentBuffer.b_cursorLine := 1;
    CurrentBuffer.b_currentLine := CurrentBuffer.b_lines;
    CurrentBuffer.b_cursorColumn := 0;
    if needRedraw then
	displayWindow();
    fi;
corp;

/*
 * cursorDown - move the cursor down one line.
 */

proc cursorDown()void:

    fixLine();		/* keep here so cursorRight is consistent */
    if CurrentBuffer.b_cursorLine < CurrentBuffer.b_lineCount then
	CurrentBuffer.b_cursorLine := CurrentBuffer.b_cursorLine + 1;
	CurrentBuffer.b_currentLine := CurrentBuffer.b_currentLine*.l_next;
	if CurrentBuffer.b_windowLine = CurrentBuffer.b_windowHeight - 1 then
	    CRT_ScrollRegionUp(CurrentBuffer.b_windowPos,
			       CurrentBuffer.b_windowHeight, 1);
	    CRT_Move(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,0);
	    displayString(CurrentBuffer.b_currentLine*.l_data);
	else
	    CurrentBuffer.b_windowLine := CurrentBuffer.b_windowLine + 1;
	fi;
    fi;
corp;

/*
 * cursorUp - move the cursor up one line if possible.
 */

proc cursorUp()void:

    fixLine();		/* keep here so cursorLeft is consistent */
    if CurrentBuffer.b_cursorLine ~= 1 then
	CurrentBuffer.b_cursorLine := CurrentBuffer.b_cursorLine - 1;
	CurrentBuffer.b_currentLine := CurrentBuffer.b_currentLine*.l_prev;
	if CurrentBuffer.b_windowLine = 0 then
	    CRT_ScrollRegionDown(CurrentBuffer.b_windowPos,
				 CurrentBuffer.b_windowHeight, 1);
	    CRT_Move(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,0);
	    displayString(CurrentBuffer.b_currentLine*.l_data);
	else
	    CurrentBuffer.b_windowLine := CurrentBuffer.b_windowLine - 1;
	fi;
    fi;
corp;

/*
 * scrollUp - scroll the screen up one line.
 */

proc scrollUp()void:
    register *LINELIST l;
    register ushort i;

    fixLine();
    if CurrentBuffer.b_cursorLine - CurrentBuffer.b_windowLine ~=
	    CurrentBuffer.b_lineCount then
	if CurrentBuffer.b_windowLine = 0 then
	    cursorDown();
	fi;
	if CurrentBuffer.b_windowHeight ~= 1 then
	    CRT_ScrollRegionUp(CurrentBuffer.b_windowPos,
			       CurrentBuffer.b_windowHeight, 1);
	    CurrentBuffer.b_windowLine := CurrentBuffer.b_windowLine - 1;
	    i := CurrentBuffer.b_windowHeight - CurrentBuffer.b_windowLine - 1;
	    if CurrentBuffer.b_cursorLine + i <= CurrentBuffer.b_lineCount then
		l := CurrentBuffer.b_currentLine;
		while i ~= 0 do
		    i := i - 1;
		    l := l*.l_next;
		od;
		CRT_Move(CurrentBuffer.b_windowPos +
			 CurrentBuffer.b_windowHeight - 1, 0);
		displayString(l*.l_data);
	    fi;
	fi;
    fi;
corp;

/*
 * scrollDown - scroll the screen down one line.
 */

proc scrollDown()void:
    register *LINELIST l;
    register ushort i;

    fixLine();
    if CurrentBuffer.b_cursorLine - CurrentBuffer.b_windowLine ~= 1 then
	if CurrentBuffer.b_windowLine = CurrentBuffer.b_windowHeight - 1 then
	    cursorUp();
	fi;
	if CurrentBuffer.b_windowHeight ~= 1 then
	    CRT_ScrollRegionDown(CurrentBuffer.b_windowPos,
				 CurrentBuffer.b_windowHeight, 1);
	    CurrentBuffer.b_windowLine := CurrentBuffer.b_windowLine + 1;
	    i := CurrentBuffer.b_windowLine;
	    l := CurrentBuffer.b_currentLine;
	    while i ~= 0 do
		i := i - 1;
		l := l*.l_prev;
	    od;
	    CRT_Move(CurrentBuffer.b_windowPos, 0);
	    displayString(l*.l_data);
	fi;
    fi;
corp;

/*
 * cursorLeft - move the cursor left one character.
 */

proc cursorLeft()void:

    if CurrentBuffer.b_cursorColumn ~= 0 or
	    CurrentBuffer.b_cursorLine ~= 1 then
	if CurrentBuffer.b_cursorColumn = 0 then
	    /* wrap to end of previous line. */
	    cursorUp();
	    unFixLine();
	    CurrentBuffer.b_cursorColumn := LineLength;
	    CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
	else
	    backUpOne();
	fi;
    fi;
corp;

/*
 * cursorRight - move the cursor right one character.
 */

proc cursorRight()void:

    if CurrentBuffer.b_cursorColumn = NCOLUMNS - 1 then
	cursorDown();
	CurrentBuffer.b_cursorColumn := 0;
    else
	forwardOne();
    fi;
corp;

/*
 * pageForward - window forward 1 screenful (keep 2 old lines).
 */

proc pageForward()void:
    register ushort i;

    if CurrentBuffer.b_cursorLine < CurrentBuffer.b_lineCount then
	fixLine();
	i := CurrentBuffer.b_windowHeight - 2;
	while CurrentBuffer.b_cursorLine ~= CurrentBuffer.b_lineCount and
		i ~= 0 do
	    i := i - 1;
	    CurrentBuffer.b_cursorLine := CurrentBuffer.b_cursorLine + 1;
	    CurrentBuffer.b_currentLine :=
		    CurrentBuffer.b_currentLine*.l_next;
	od;
	displayWindow();
    fi;
corp;

/*
 * pageBackward - window backward 1 screenful (keep 2 old lines).
 */

proc pageBackward()void:
    register ushort i;

    if CurrentBuffer.b_cursorLine ~= 1 then
	fixLine();
	i := CurrentBuffer.b_windowHeight - 2;
	while CurrentBuffer.b_cursorLine ~= 1 and i ~= 0 do
	    i := i - 1;
	    CurrentBuffer.b_currentLine :=
		    CurrentBuffer.b_currentLine*.l_prev;
	    CurrentBuffer.b_cursorLine := CurrentBuffer.b_cursorLine - 1;
	    if CurrentBuffer.b_cursorLine - 1 <
		    CurrentBuffer.b_windowLine then
		CurrentBuffer.b_windowLine := CurrentBuffer.b_windowLine - 1;
	    fi;
	od;
	displayWindow();
    fi;
corp;

/*
 * skipWord - move cursor forward one word.
 */

proc skipWord()void:

    if isLetDig() then
	while CursorPos* ~= '\e' and isLetDig() do
	    forwardOne();
	od;
    elif CursorPos* ~= '\e' then
	forwardOne();
    fi;
    while CursorPos* ~= '\e' and isBlank() do
	forwardOne();
    od;
corp;

/*
 * backWord - move cursor back one word.
 */

proc backWord()void:

    if CurrentBuffer.b_cursorColumn ~= 0 then
	backUpOne();
    fi;
    while CurrentBuffer.b_cursorColumn ~= 0 and isBlank() do
	backUpOne();
    od;
    if CurrentBuffer.b_cursorColumn ~= 0 and isLetDig() then
	while CurrentBuffer.b_cursorColumn ~= 0 and isLetDig() do
	    backUpOne();
	od;
	if CursorPos* ~= '\e' and not isLetDig() then
	    forwardOne();
	fi;
    fi;
corp;

/*
 * gotoLine - go to a specific line in the file.
 */

proc gotoLine()void:
    register ulong n;
    bool needRedraw;

    fixLine();
    n := getNumber("Go to which line: ");
    if n > CurrentBuffer.b_lineCount then
	n := CurrentBuffer.b_lineCount;
    fi;
    if n ~= 0 then
	needRedraw :=
	    if n >= CurrentBuffer.b_cursorLine-CurrentBuffer.b_windowLine and
		    n < CurrentBuffer.b_cursorLine -
			CurrentBuffer.b_windowLine +
			CurrentBuffer.b_windowHeight then
		CurrentBuffer.b_windowLine := CurrentBuffer.b_windowLine +
			make(n - CurrentBuffer.b_cursorLine, ushort);
		false
	    else
		true
	    fi;
	CurrentBuffer.b_cursorLine := n;
	CurrentBuffer.b_currentLine := CurrentBuffer.b_lines;
	while
	    n := n - 1;
	    n ~= 0
	do
	    CurrentBuffer.b_currentLine :=
		    CurrentBuffer.b_currentLine*.l_next;
	od;
	if needRedraw then
	    displayNew();
	fi;
    fi;
corp;

/***************************************************************************/

/*
 * ROUTINES INVOLVING SEARCHING AND REPLACING
 *
 * Note the use of CurrentBuffer.b_windowLine as a sort of flag controlling
 * the display of search targets - if it's equal to the window height, then
 * the target was out of the window and a redisplay is needed.
 */

/*
 * findForward - find the next occurrence of Subject in forward direction.
 */

proc findForward()bool:
    register *char p2, p3;
    bool match;

    unFixLine();
    fixPosition();
    match := false;
    while not match and
	    (CurrentBuffer.b_cursorLine < CurrentBuffer.b_lineCount or
	     CursorPos* ~= '\e') do
	if CursorPos* = '\e' then
	    fixLine();
	    CurrentBuffer.b_currentLine := CurrentBuffer.b_currentLine*.l_next;
	    CurrentBuffer.b_cursorLine := CurrentBuffer.b_cursorLine + 1;
	    CurrentBuffer.b_cursorColumn := 0;
	    if CurrentBuffer.b_windowLine ~= CurrentBuffer.b_windowHeight then
		CurrentBuffer.b_windowLine := CurrentBuffer.b_windowLine + 1;
	    fi;
	    unFixLine();
	    if LineLength < SubjectLength then
		CurrentBuffer.b_cursorColumn := LineLength;
		CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
	    fi;
	else
	    p2 := &Subject[0];
	    p3 := CursorPos;
	    while p3* = p2* and p2* ~= '\e' do
		p2 := p2 + 1;
		p3 := p3 + 1;
	    od;
	    if p2* = '\e' then
		match := true;
	    else
		forwardOne();
	    fi;
	fi;
    od;
    match
corp;

/*
 * findReverse - find the next occurrence of Subject in reverse direction.
 */

proc findReverse()bool:
    register *char p2, p3;
    bool match, checkedFirst;

    match := false;
    checkedFirst := false;
    unFixLine();
    fixPosition();
    if LineLength < SubjectLength then
	CurrentBuffer.b_cursorColumn := 0;
	CursorPos := &LineBuffer[0];
	checkedFirst := true;
    fi;
    while not match and
	    (CurrentBuffer.b_cursorLine >= 1 or
	     CurrentBuffer.b_cursorColumn > 0 or not checkedFirst) do
	if CurrentBuffer.b_cursorColumn = 0 and checkedFirst then
	    fixLine();
	    CurrentBuffer.b_currentLine := CurrentBuffer.b_currentLine*.l_prev;
	    CurrentBuffer.b_cursorLine := CurrentBuffer.b_cursorLine - 1;
	    unFixLine();
	    if LineLength < SubjectLength then
		CurrentBuffer.b_cursorColumn := 0;
		CursorPos := &LineBuffer[0];
	    else
		checkedFirst := false;
		CurrentBuffer.b_cursorColumn := LineLength - SubjectLength;
		CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
	    fi;
	    if CurrentBuffer.b_windowLine ~=
		    CurrentBuffer.b_windowHeight then
		CurrentBuffer.b_windowLine :=
		    if CurrentBuffer.b_windowLine = 0 then
			CurrentBuffer.b_windowHeight
		    else
			CurrentBuffer.b_windowLine - 1
		    fi;
	    fi;
	else
	    if CurrentBuffer.b_cursorColumn = 0 then
		checkedFirst := true;
	    fi;
	    p2 := &Subject[0];
	    p3 := CursorPos;
	    while p3* = p2* and p2* ~= '\e' do
		p2 := p2 + 1;
		p3 := p3 + 1;
	    od;
	    if p2* = '\e' then
		match := true;
	    else
		if CurrentBuffer.b_cursorColumn ~= 0 then
		    backUpOne();
		fi;
	    fi;
	fi;
    od;
    if CurrentBuffer.b_cursorLine = 0 then
	CurrentBuffer.b_cursorLine := 1;
	CurrentBuffer.b_currentLine := CurrentBuffer.b_lines;
    fi;
    match
corp;

/*
 * noMatch - say No match.
 */

proc noMatch()void:

    putError("No match.");
corp;

/*
 * globalChange - go through and make a global change.
 */

proc globalChange()void:
    [SEARCHLENGTH] char firstHalf @ LineBuffer;
    register *LINELIST redrawPtr;
    register ulong found;
    register ushort redrawLine;
    bool hadTooLong, needRedraw;

    getString("Change: ");
    if LineBuffer[0] ~= '\e' then
	Searching := true;
	Subject := firstHalf;
	getString("To: ");
	Object := firstHalf;
	putError("Changing...");
	CRT_Flush();
	SubjectLength := CharsLen(&Subject[0]);
	ToLength := CharsLen(&Object[0]);
	needRedraw := false;
	hadTooLong := false;
	redrawPtr := nil;
	found := 0;
	while findForward() do
	    found := found + 1;
	    if LineLength - SubjectLength + ToLength >= NCOLUMNS then
		hadTooLong := true;
		/* skip past the found text, so we don't refind it */
		CursorPos := CursorPos + SubjectLength;
		CurrentBuffer.b_cursorColumn :=
		    SubjectLength + CurrentBuffer.b_cursorColumn;
	    else
		/* redraw any line which we have changed and which is
		   in the displayed window and which isn't the one we
		   are still looking at */
		if redrawPtr ~= nil and
			redrawPtr ~= CurrentBuffer.b_currentLine then
		    CRT_ClearLine(CurrentBuffer.b_windowPos + redrawLine);
		    displayString(redrawPtr*.l_data);
		    CRT_Flush();
		    redrawPtr := nil;
		fi;
		/* we have changed THIS line - set up to redraw it */
		if CurrentBuffer.b_windowLine ~=
			CurrentBuffer.b_windowHeight then
		    redrawLine := CurrentBuffer.b_windowLine;
		    redrawPtr := CurrentBuffer.b_currentLine;
		else
		    /* we have gone beyond the visible window - set a flag
		       saying that we need to redraw when we're done */
		    needRedraw := true;
		fi;
		/* do the actual change; don't redisplay now */
		deleteChunk(CursorPos, CursorPos + SubjectLength, false);
		insertChunk(&Object[0], false);
	    fi;
	od;
	fixLine();
	if needRedraw or CurrentBuffer.b_windowLine =
		CurrentBuffer.b_windowHeight then
	    displayNew();
	else
	    if redrawLine ~= CurrentBuffer.b_windowHeight then
		CRT_ClearLine(CurrentBuffer.b_windowPos + redrawLine);
		displayString(redrawPtr*.l_data);
	    fi;
	fi;
	clearError();
	if hadTooLong then
	    tooLong();
	elif found = 0 then
	    noMatch();
	elif not NoMemory and not TooLong then
	    clearError();
	    putError("Found ");
	    putNumber(found, 4);
	    CRT_PutChars(" match(es).");
	fi;
    fi;
corp;

/*
 * displayFind - display the next match found.
 */

proc displayFind()void:

    if CurrentBuffer.b_windowLine = CurrentBuffer.b_windowHeight then
	/* found subject is out of current screenful. */
	displayNew();
    fi;
corp;

/*
 * queryReplace - enter query/replace mode.
 */

proc queryReplace()void:
    [SEARCHLENGTH] char firstHalf @ LineBuffer;
    char ch;
    bool quit;

    getString("Change: ");
    if LineBuffer[0] ~= '\e' then
	Searching := true;
	Subject := firstHalf;
	getString("To: ");
	Object := firstHalf;
	SubjectLength := CharsLen(&Subject[0]);
	ToLength := CharsLen(&Object[0]);
	quit := false;
	while not quit do
	    putError("Searching...");
	    CRT_Flush();
	    if findForward() then
		fixLine();
		clearError();
		if LineLength - SubjectLength + ToLength >= NCOLUMNS then
		    putError("Can't change - line too long: ");
		    displayFind();
		    update();
		    ch := CRT_GetChar();
		    forwardOne();
		else
		    displayFind();
		    CRT_ClearLine(NLines - 1);
		    CRT_PutChars("Change? (Y/N/Q) ");
		    update();
		    ch := CRT_GetChar();
		    if ch = 'y' or ch = 'Y' then
			unFixLine();
			deleteChunk(CursorPos, CursorPos + SubjectLength,true);
			update();
			insertChunk(&Object[0], true);
		    elif ch = 'n' or ch = 'N' then
			forwardOne();
		    elif ch = 'q' or ch = 'Q' then
			quit := true;
		    else
			CRT_PutChar('\(BEEP)');
			/* findForward will find the same one again */
		    fi;
		fi;
	    else
		fixLine();
		clearError();
		displayFind();
		noMatch();
		quit := true;
	    fi;
	od;
	clearError();
    fi;
corp;

/*
 * find - start a search in the forward direction.
 */

proc find()void:
    [SEARCHLENGTH] char firstHalf @ LineBuffer;
    ulong saveLine;
    ushort saveColumn;

    getString("Find: ");
    if LineBuffer[0] ~= '\e' then
	Searching := true;
	Subject := firstHalf;
	SubjectLength := CharsLen(&Subject[0]);
    fi;
    if Searching then
	putError("Searching...");
	CRT_Flush();
	saveLine := CurrentBuffer.b_cursorLine;
	saveColumn := CurrentBuffer.b_cursorColumn;
	if findForward() then
	    if CurrentBuffer.b_cursorLine = saveLine and
		    CurrentBuffer.b_cursorColumn = saveColumn then
		forwardOne();
		if findForward() then
		    clearError();
		    displayFind();
		else
		    clearError();
		    displayFind();
		    noMatch();
		fi;
	    else
		clearError();
		displayFind();
	    fi;
	else
	    clearError();
	    displayFind();
	    noMatch();
	fi;
    fi;
corp;

/*
 * reverseFind - start a search in the reverse direction.
 */

proc reverseFind()void:
    [SEARCHLENGTH] char firstHalf @ LineBuffer;
    ulong saveLine;
    ushort saveColumn;

    getString("Reverse find: ");
    if LineBuffer[0] ~= '\e' then
	Searching := true;
	Subject := firstHalf;
	SubjectLength := CharsLen(&Subject[0]);
    fi;
    if Searching then
	putError("Searching...");
	CRT_Flush();
	saveLine := CurrentBuffer.b_cursorLine;
	saveColumn := CurrentBuffer.b_cursorColumn;
	if findReverse() and (saveLine ~= 1 or saveColumn ~= 0) then
	    if CurrentBuffer.b_cursorLine = saveLine and
		    CurrentBuffer.b_cursorColumn = saveColumn then
		if saveColumn = 0 then
		    fixLine();
		    CurrentBuffer.b_cursorLine := saveLine - 1;
		    CurrentBuffer.b_currentLine :=
			CurrentBuffer.b_currentLine*.l_prev;
		    unFixLine();
		    if LineLength < SubjectLength then
			CurrentBuffer.b_cursorColumn := 0;
			CursorPos := &LineBuffer[0];
		    else
			CurrentBuffer.b_cursorColumn :=
				LineLength - SubjectLength;
			CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
		    fi;
		    if CurrentBuffer.b_windowLine ~=
			    CurrentBuffer.b_windowHeight then
			CurrentBuffer.b_windowLine :=
			    if CurrentBuffer.b_windowLine = 0 then
				CurrentBuffer.b_windowHeight
			    else
				CurrentBuffer.b_windowLine - 1
			    fi;
		    fi;
		else
		    backUpOne();
		fi;
		if findReverse() then
		    clearError();
		    displayFind();
		else
		    clearError();
		    displayFind();
		    noMatch();
		fi;
	    else
		clearError();
		displayFind();
	    fi;
	else
	    clearError();
	    displayFind();
	    noMatch();
	fi;
    fi;
corp;

/***************************************************************************/

/*
 * MISCELLANEOUS EDITING ROUTINES
 */

/*
 * deleteChar - the delete character command, or the format command is
 *		operating in delete mode.
 */

proc deleteChar()void:

    unFixLine();
    fixPosition();
    if CurrentBuffer.b_cursorColumn ~= LineLength then
	deleteChunk(CursorPos, CursorPos + 1, true);
    fi;
corp;

/*
 * tabKey - effectively put a tab character in.
 */

proc tabKey()void:
    register ushort column;

    unFixLine();
    fixPosition();
    if not CurrentBuffer.b_insertMode then
	if CurrentBuffer.b_cursorColumn ~= LineLength then
	    deleteChunk(CursorPos, CursorPos + 1, true);
	    update();
	fi;
    fi;
    column := CurrentBuffer.b_cursorColumn + 1;
    while column < NCOLUMNS and not TabStops[column] do
	column := column + 1;
    od;
    if column > NCOLUMNS - 1 then
	column := CurrentBuffer.b_cursorColumn + 1;
	if column >= NCOLUMNS then
	    column := column - 1;
	fi;
    fi;
    insertBlanks(column - CurrentBuffer.b_cursorColumn, true);
corp;

/*
 * deleteNextWord - delete word in forward direction.
 */

proc deleteNextWord()void:
    register *char startPos;
    ushort column;

    unFixLine();
    fixPosition();
    if LineLength ~= CurrentBuffer.b_cursorColumn then
	column := CurrentBuffer.b_cursorColumn;
	startPos := CursorPos;
	skipWord();
	deleteChunk(startPos, CursorPos, true);
	CurrentBuffer.b_cursorColumn := column;
	CursorPos := startPos;
    fi;
corp;

/*
 * deletePrevWord - delete word in backwards direction.
 */

proc deletePrevWord()void:
    *char startPos;

    unFixLine();
    fixPosition();
    if CurrentBuffer.b_cursorColumn ~= 0 then
	startPos := CursorPos;
	backWord();
	deleteChunk(CursorPos, startPos, true);
    fi;
corp;

/*
 * deleteToEnd - delete from current position to end of line.
 */

proc deleteToEnd()void:

    unFixLine();
    fixPosition();
    CursorPos* := '\e';
    LineChanged := true;
    LineLength := CurrentBuffer.b_cursorColumn;
    CRT_ClearTail();
corp;

/*
 * setPoint - set a first end here.
 */

proc setPoint()void:

    if CurrentBuffer.b_currentLine ~= nil then
	if CurrentBuffer.b_gotRegion and
		CurrentBuffer.b_regionLine = CurrentBuffer.b_cursorLine and
		CurrentBuffer.b_regionColumn = CurrentBuffer.b_cursorColumn and
		CurrentBuffer.b_regionPtr = CurrentBuffer.b_currentLine and
		CurrentBuffer.b_regionPos = CursorPos then
	    CurrentBuffer.b_gotRegion := false;
	    regionStatus();
	else
	    if not CurrentBuffer.b_gotRegion then
		CurrentBuffer.b_gotRegion := true;
		regionStatus();
	    fi;
	    unFixLine();
	    fixPosition();
	    CurrentBuffer.b_regionLine := CurrentBuffer.b_cursorLine;
	    CurrentBuffer.b_regionColumn := CurrentBuffer.b_cursorColumn;
	    CurrentBuffer.b_regionPtr := CurrentBuffer.b_currentLine;
	    CurrentBuffer.b_regionPos := CursorPos;
	fi;
    fi;
corp;

/*
 * setNumeric - set a numeric parameter.
 */

proc setNumeric()void:

    case CRT_GetChar()
    incase 'i':
	setParameter("indent increment",
		     &CurrentBuffer.b_indentIncrement, 0, 70);
    incase 'l':
	setParameter("indent amount", &CurrentBuffer.b_indentAmount, 0,
		     CurrentBuffer.b_wrapColumn - 1);
    incase 'w':
	setParameter("wrap column", &CurrentBuffer.b_wrapColumn,
		     CurrentBuffer.b_indentAmount + 1, NCOLUMNS - 1);
    default:
	putError("Illegal set command.");
    esac;
corp;

/*
 * blankMode - set the mode for blank output.
 */

proc blankMode()void:
    *char
	BLANKSMESS = "Blanks used.",
	TABSMESS = "8-column tabs used.",
	COUNTSMESS = "Compressed blanks used.";

    putError(
	if CurrentBuffer.b_tabMode = BLANKS then
	    CurrentBuffer.b_tabMode := TABS;
	    TABSMESS
	elif CurrentBuffer.b_tabMode = TABS then
	    CurrentBuffer.b_tabMode := COUNTS;
	    COUNTSMESS
	else
	    CurrentBuffer.b_tabMode := BLANKS;
	    BLANKSMESS
	fi);
corp;

/*
 * insertLine - try to insert a new empty line after the current line.
 */

proc insertLine()void:
    register *LINELIST l;
    register *char p;

    fixLine();
    initLines();
    l := new(LINELIST);
    if l = nil then
	noSpace();
    else
	p := new(char);
	if p = nil then
	    free(l);
	    noSpace();
	else
	    CurrentBuffer.b_changeCount := CurrentBuffer.b_changeCount + 1;
	    if CurrentBuffer.b_gotRegion and
		    CurrentBuffer.b_regionLine >
		    CurrentBuffer.b_cursorLine then
		CurrentBuffer.b_regionLine := CurrentBuffer.b_regionLine + 1;
	    fi;
	    p* := '\e';
	    CurrentBuffer.b_lineCount := CurrentBuffer.b_lineCount + 1;
	    l*.l_next := CurrentBuffer.b_currentLine*.l_next;
	    l*.l_prev := CurrentBuffer.b_currentLine;
	    l*.l_data := p;
	    CurrentBuffer.b_currentLine*.l_next := l;
	    if l*.l_next ~= nil then
		l*.l_next*.l_prev := l;
	    fi;
	    if CurrentBuffer.b_cursorLine = CurrentBuffer.b_lineCount - 1 or
		    CurrentBuffer.b_windowLine =
			CurrentBuffer.b_windowHeight - 1 then
		/* inserting at the end of the file or on last window line */
		cursorDown();
	    else
		/* inserting in the middle of the file and window */
		cursorDown();
		CRT_ScrollRegionDown(
		    CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,
		    CurrentBuffer.b_windowHeight - CurrentBuffer.b_windowLine,
		    1);
	    fi;
	    CurrentBuffer.b_cursorColumn := 0;
	    update();
	    insertBlanks(CurrentBuffer.b_indentAmount, true);
	fi;
    fi;
corp;

/*
 * carriageReturn - the user has typed a carriage return (or the format
 *		    command has reached the end of a buffered line).
 */

proc carriageReturn()void:

    if CurrentBuffer.b_multiColumn then
	pretend(wrapMulti('\r'), void);
    else
	insertLine();
    fi;
corp;

/*
 * deleteLine - delete the current line.
 */

proc deleteLine()void:
    register *LINELIST l, l1;
    register ushort i;

    if CurrentBuffer.b_lineCount ~= 0 then
	CurrentBuffer.b_lineCount := CurrentBuffer.b_lineCount - 1;
	CurrentBuffer.b_changeCount := CurrentBuffer.b_changeCount + 1;
	if CurrentBuffer.b_gotRegion then
	    if CurrentBuffer.b_regionLine = CurrentBuffer.b_cursorLine then
		CurrentBuffer.b_gotRegion := false;
		regionStatus();
	    elif CurrentBuffer.b_regionLine > CurrentBuffer.b_cursorLine then
		CurrentBuffer.b_regionLine := CurrentBuffer.b_regionLine - 1;
	    fi;
	fi;
	fixLine();
	l := CurrentBuffer.b_currentLine;
	if l*.l_prev = nil then
	    CurrentBuffer.b_lines := l*.l_next;
	else
	    l*.l_prev*.l_next := l*.l_next;
	fi;
	if l*.l_next ~= nil then
	    l*.l_next*.l_prev := l*.l_prev;
	    CurrentBuffer.b_currentLine := l*.l_next;
	    i := CurrentBuffer.b_windowHeight - CurrentBuffer.b_windowLine - 1;
	    if i ~= 0 then
		CRT_ScrollRegionUp(
		    CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,
		    i + 1, 1);
		l1 := CurrentBuffer.b_currentLine;
		while l1 ~= nil and i ~= 0 do
		    i := i - 1;
		    l1 := l1*.l_next;
		od;
		if l1 ~= nil then
		    CRT_Move(CurrentBuffer.b_windowPos +
			     CurrentBuffer.b_windowHeight - 1, 0);
		    displayString(l1*.l_data);
		fi;
	    else
		CRT_ClearLine(CurrentBuffer.b_windowPos +
			      CurrentBuffer.b_windowLine);
		displayString(CurrentBuffer.b_currentLine*.l_data);
	    fi;
	else
	    /* last line in file deleted */
	    CRT_ClearLine(CurrentBuffer.b_windowPos +
			  CurrentBuffer.b_windowLine);
	    cursorUp();
	fi;
	freeString(l*.l_data);
	free(l);
    fi;
corp;

/*
 * joinLines - join the next line to the end of the current line.
 */

proc joinLines()void:
    register *char p;
    register ushort len;
    bool wantUp;

    if CurrentBuffer.b_cursorLine < CurrentBuffer.b_lineCount then
	fixLine();
	CurrentBuffer.b_currentLine := CurrentBuffer.b_currentLine*.l_next;
	CurrentBuffer.b_cursorColumn := 0;
	unFixLine();
	len := LineLength;
	/* trim all but one leading blank from the second line */
	while CursorPos* = ' ' do
	    forwardOne();
	    len := len - 1;
	od;
	if CurrentBuffer.b_cursorColumn ~= 0 then
	    backUpOne();
	    len := len + 1;
	fi;
	p := pretend(Malloc(len + 1), *char);
	if p = nil then
	    noSpace();
	    fixLine();
	    CurrentBuffer.b_currentLine := CurrentBuffer.b_currentLine*.l_prev;
	else
	    CharsCopy(p, CursorPos);
	    fixLine();
	    CurrentBuffer.b_currentLine := CurrentBuffer.b_currentLine*.l_prev;
	    unFixLine();
	    if LineLength + len > NCOLUMNS - 1 then
		tooLong();
	    else
		CurrentBuffer.b_cursorColumn := LineLength;
		CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
		update();
		insertChunk(p, true);
		cursorDown();
		wantUp := CurrentBuffer.b_currentLine*.l_next ~= nil;
		deleteLine();
		if wantUp then
		    cursorUp();
		fi;
	    fi;
	    freeString(p);
	fi;
    fi;
corp;

/*
 * splitLine - split a line into two lines at the current position.
 */

proc splitLine()void:
    register *char p;

    unFixLine();
    fixPosition();
    if LineLength - CurrentBuffer.b_cursorColumn +
		CurrentBuffer.b_indentAmount >= NCOLUMNS then
	tooLong();
    else
	p := pretend(Malloc(CharsLen(CursorPos) + 1), *char);
	if p = nil then
	    noSpace();
	else
	    CharsCopy(p, CursorPos);
	    update();
	    CRT_ClearTail();
	    CursorPos* := '\e';
	    LineChanged := true;
	    insertLine();
	    update();
	    insertChunk(p, true);
	    freeString(p);
	    fixLine();
	fi;
    fi;
corp;

/*
 * insertArb - insert an arbitrary character into the current line.
 */

proc insertArb()void:
    [2] char buffer;
    char ch;

    fixLine();
    ch := getNumber("Decimal value: ") + '\e';
    if ch ~= '\e' then
	if ch = '\r' or ch = '\n' then
	    CRT_PutChar(BEEP);
	else
	    buffer[0] := ' ';
	    buffer[1] := '\e';
	    update();
	    insertChunk(&buffer[0], true);
	    if not TooLong then
		(CursorPos - 1)* := ch;
		update();
		CRT_PutChar('\b');
		displayChar(ch);
	    fi;
	fi;
    fi;
corp;

/*
 * adjustLine - adjust the current line so the first nonblank is at the
 *		current left indent position.
 */

proc adjustLine(bool needRedraw)void:
    register ushort amount;

    unFixLine();
    CurrentBuffer.b_cursorColumn := 0;
    CursorPos := &LineBuffer[0];
    while CursorPos* = ' ' do
	forwardOne();
    od;
    if LineLength + CurrentBuffer.b_indentAmount -
		CurrentBuffer.b_cursorColumn >= NCOLUMNS then
	tooLong();
    else
	if CurrentBuffer.b_indentAmount > CurrentBuffer.b_cursorColumn then
	    /* adding indentation */
	    insertBlanks(
		CurrentBuffer.b_indentAmount - CurrentBuffer.b_cursorColumn,
		needRedraw);
	else
	    /* deleting indentation */
	    amount :=
		CurrentBuffer.b_cursorColumn - CurrentBuffer.b_indentAmount;
	    deleteChunk(&LineBuffer[0], &LineBuffer[amount], needRedraw);
	    CurrentBuffer.b_cursorColumn :=
		CurrentBuffer.b_cursorColumn - amount;
	    CursorPos := CursorPos - amount;
	fi;
    fi;
corp;

/***************************************************************************/

/*
 * SOME WINDOW FIDDLING ROUTINES.
 */

/*
 * getOtherName - get the name of a buffer other than the current.
 */

proc getOtherName()*BUFFERLIST:
    register *BUFFERLIST b;
    *BUFFERLIST match, first;
    register *char p, q;
    ushort matchCount;

    getString("Buffer: ");
    if LineBuffer[0] ~= '\e' then
	BufferPtr* := CurrentBuffer;
	matchCount := 0;
	b := BufferList;
	first := b;
	while
	    p := &LineBuffer[0];
	    q := b*.b_fileName;
	    while p* ~= '\e' and p* = q* do
		p := p + 1;
		q := q + 1;
	    od;
	    if p* = '\e' then
		match := b;
		matchCount := matchCount + 1;
	    fi;
	    b := b*.b_next;
	    b ~= first
	do
	od;
	if matchCount = 0 then
	    putError("no such buffer");
	    nil
	elif matchCount ~= 1 then
	    putError("ambiguous");
	    nil
	elif match = BufferPtr then
	    putError("that's this one!");
	    nil
	else
	    match
	fi
    else
	nil
    fi
corp;

/*
 * deleteWindow - delete the current window from the screen.
 */

proc deleteWindow()void:
    register *BUFFERLIST old;
    ushort oldPos, extraHeight, oldHeight;

    fixLine();
    old := BufferPtr;
    old* := CurrentBuffer;
    extraHeight := CurrentBuffer.b_windowHeight + 1;
    oldPos := CurrentBuffer.b_windowPos;
    if CurrentBuffer.b_windowPos ~= 0 then
	/* extend the previous window */
	BufferPtr := CurrentBuffer.b_sPrev;
	old*.b_sNext*.b_sPrev := BufferPtr;	/* BEFORE fetch struct!! */
	CurrentBuffer := BufferPtr*;
	CurrentBuffer.b_sNext := old*.b_sNext;
	oldHeight := CurrentBuffer.b_windowHeight;
	CurrentBuffer.b_windowHeight := oldHeight + extraHeight;
	ScreenList := BufferPtr;
	old*.b_sNext := nil;
	old*.b_sPrev := nil;
	displayBottom(oldHeight);
	newBuffer();
	fullStatus();
    elif CurrentBuffer.b_windowPos + CurrentBuffer.b_windowHeight ~=
	    NLines - 2 then
	/* extend the next window */
	BufferPtr := CurrentBuffer.b_sNext;
	old*.b_sPrev*.b_sNext := BufferPtr;	/* BEFORE fetch struct!! */
	CurrentBuffer := BufferPtr*;
	CurrentBuffer.b_sPrev := old*.b_sPrev;
	oldHeight := CurrentBuffer.b_windowHeight;
	CurrentBuffer.b_windowHeight := oldHeight + extraHeight;
	CurrentBuffer.b_windowPos := oldPos;
	CRT_ScrollRegionUp(CurrentBuffer.b_windowPos,
		CurrentBuffer.b_windowHeight, extraHeight);
	ScreenList := BufferPtr;
	old*.b_sNext := nil;
	old*.b_sPrev := nil;
	displayBottom(oldHeight);
	newBuffer();
	fullStatus();
    /* we do nothing if it's the only window on the screen */
    fi;
corp;

/*
 * oneWindow - make the current window the only window.
 */

proc oneWindow()void:
    register *BUFFERLIST b, first, temp;
    ushort oldHeight;

    if CurrentBuffer.b_sNext ~= BufferPtr then
	fixLine();
	if CurrentBuffer.b_windowPos ~= 0 then
	    /* delete the stuff above this window */
	    CRT_ScrollRegionUp(0,
		CurrentBuffer.b_windowPos + CurrentBuffer.b_windowHeight,
		CurrentBuffer.b_windowPos);
	fi;
	oldHeight := CurrentBuffer.b_windowHeight;
	b := ScreenList;
	first := b;
	while
	    temp := b*.b_sNext;
	    if temp = nil then
		/* oh-oh, that damn bug again!! */
		putError("oneWindow bug!!!!");
		writeln("This: ", b*.b_fileName);
		writeln("Prev: ", b*.b_sPrev*.b_fileName);
		false
	    else
		b*.b_sNext := nil;
		b*.b_sPrev := nil;
		b := temp;
		b ~= first
	    fi
	do
	od;
	CurrentBuffer.b_sNext := BufferPtr;
	CurrentBuffer.b_sPrev := BufferPtr;
	ScreenList := BufferPtr;
	CurrentBuffer.b_windowPos := 0;
	CurrentBuffer.b_windowHeight := NLines - 2;
	displayBottom(oldHeight);
	fullStatus();
    fi;
corp;

/*
 * makeWindow - make a window for a new/other buffer (as CurrentBuffer).
 */

proc makeWindow()void:
    register *BUFFERLIST b, first, selected;
    ulong longestTime;
    ushort maxHeight, height1, i;

    /* algorithm: if there is a window on the screen that is big enough to
       split into two, then we do so, and we use the bottom half of it for
       the new window, otherwise, the least recently used window is reused
       as our new window. */

    if ScreenList = nil then
	CurrentBuffer.b_sPrev := BufferPtr;
	CurrentBuffer.b_sNext := BufferPtr;
	ScreenList := BufferPtr;
	CurrentBuffer.b_windowPos := 0;
	CurrentBuffer.b_windowHeight := NLines - 2;
    else
	maxHeight := 0;
	b := ScreenList;
	first := b;
	while
	    if b*.b_windowHeight > maxHeight then
		maxHeight := b*.b_windowHeight;
		selected := b;
	    fi;
	    b := b*.b_sNext;
	    b ~= first
	do
	od;
	if maxHeight > MINIMUM_WINDOW_HEIGHT * 2 then
	    /* we can split the selected window */
	    CurrentBuffer.b_sPrev := selected;
	    CurrentBuffer.b_sNext := selected*.b_sNext;
	    CurrentBuffer.b_sNext*.b_sPrev := BufferPtr;
	    selected*.b_sNext := BufferPtr;
	    height1 := (maxHeight - 1) / 2;
	    CurrentBuffer.b_windowHeight := maxHeight - 1 - height1;
	    CurrentBuffer.b_windowPos := selected*.b_windowPos + height1 + 1;
	    selected*.b_windowHeight := height1;
	    if selected*.b_windowLine >= height1 then
		/* be nice - scroll window to put it's current line visible */
		i := selected*.b_windowLine - height1;
		CRT_ScrollRegionUp(selected*.b_windowPos, maxHeight, i + 1);
		selected*.b_windowLine := height1 - 1;
	    fi;
	    BufferPtr* := CurrentBuffer;
	    BufferPtr := CurrentBuffer.b_sPrev;
	    CurrentBuffer := BufferPtr*;
	    fullStatus();
	    BufferPtr := CurrentBuffer.b_sNext;
	    CurrentBuffer := BufferPtr*;
	else
	    /* we re-use the least recently used window */
	    longestTime := 0xffffffff;
	    b := ScreenList;
	    first := b;
	    while
		if b*.b_LRU < longestTime then
		    longestTime := b*.b_LRU;
		    selected := b;
		fi;
		b := b*.b_sNext;
		b ~= first
	    do
	    od;
	    CurrentBuffer.b_windowPos := selected*.b_windowPos;
	    CurrentBuffer.b_windowHeight := selected*.b_windowHeight;
	    CurrentBuffer.b_sPrev := selected*.b_sPrev;
	    CurrentBuffer.b_sNext := selected*.b_sNext;
	    ScreenList := BufferPtr;
	    CurrentBuffer.b_sPrev*.b_sNext := BufferPtr;
	    CurrentBuffer.b_sNext*.b_sPrev := BufferPtr;
	    selected*.b_sNext := nil;
	    selected*.b_sPrev := nil;
	fi;
    fi;
corp;

/*
 * splitWindow - split the current window into two windows.
 */

proc splitWindow()void:
    register *BUFFERLIST b;

    fixLine();
    b := getOtherName();
    if b ~= nil then
	BufferPtr* := CurrentBuffer;
	BufferPtr := b;
	CurrentBuffer := b*;
	makeWindow();
	displayWindow();
	fullStatus();
	newBuffer();
    fi;
corp;

/***************************************************************************/

/*
 * BUFFER FIDDLING ROUTINES.
 */

/*
 * editNew - editing another file - get set up and go with it.
 */

proc editNew(*char fileName)bool:
    register *BUFFERLIST b;

    b := new(BUFFERLIST);
    if b = nil then
	noSpace();
	false
    else
	if BufferList ~= nil then
	    BufferPtr* := CurrentBuffer;
	    b*.b_next := BufferList;
	    b*.b_prev := BufferList*.b_prev;
	    b*.b_prev*.b_next := b;
	    BufferList*.b_prev := b;
	else
	    b*.b_next := b;
	    b*.b_prev := b;
	    BufferList := b;
	fi;
	BufferPtr := b;
	CurrentBuffer := b*;
	makeWindow();
	newBuffer();
	if editFile(fileName) then
	    true
	else
	    if b*.b_next = b then
		BufferList := nil;
		ScreenList := nil;
		CRT_ClearScreen();
	    else
		deleteWindow();
		BufferPtr := BufferList;
		BufferPtr*.b_prev := b*.b_prev;
		b*.b_prev*.b_next := BufferPtr;
		CurrentBuffer := BufferPtr*;
	    fi;
	    free(b);
	    false
	fi
    fi
corp;

/*
 * visitFile - read in a new file.
 *	       NOTE: for now we ALWAYS read it in again.
 */

proc visitFile()void:

    fixLine();
    if getFileName("Visit file: ") then
	pretend(editNew(&LineBuffer[0]), void);
    fi;
corp;

/*
 * deleteBuffer - delete the current buffer.
 */

proc deleteBuffer()void:
    register *BUFFERLIST b;

    fixLine();
    if checkAbandon() then
	b := BufferPtr;
	if CurrentBuffer.b_next = b then
	    /* it's the only buffer we have - go back to having none */
	    ScreenList := nil;
	    BufferList := nil;
	    CRT_ClearScreen();
	elif CurrentBuffer.b_sNext = b then
	    /* it's the only one on the screen - move in another */
	    BufferPtr := CurrentBuffer.b_next;
	    BufferPtr*.b_prev := CurrentBuffer.b_prev;
	    CurrentBuffer.b_prev*.b_next := BufferPtr;
	    CurrentBuffer := BufferPtr*;
	    ScreenList := BufferPtr;
	    if BufferList = b then
		BufferList := b*.b_next;
	    fi;
	    CurrentBuffer.b_sNext := BufferPtr;
	    CurrentBuffer.b_sPrev := BufferPtr;
	    CurrentBuffer.b_windowPos := 0;
	    CurrentBuffer.b_windowHeight := NLines - 2;
	    displayWindow();
	    fullStatus();
	else
	    /* we just delete this guy's window */
	    deleteWindow();
	    /* deleteWindow will have changed BufferPtr and CurrentBuffer */
	    BufferPtr* := CurrentBuffer;
	    b*.b_next*.b_prev := b*.b_prev;
	    b*.b_prev*.b_next := b*.b_next;
	    CurrentBuffer := BufferPtr*;
	    if BufferList = b then
		BufferList := b*.b_next;
	    fi;
	fi;
	clearMemory(b*.b_lines);
	freeString(b*.b_fileName);
	free(b);
    fi;
corp;

/*
 * otherBuffer - switch to another buffer.
 */

proc otherBuffer()void:
    register *BUFFERLIST b;

    fixLine();
    BufferPtr* := CurrentBuffer;
    b := getOtherName();
    if b ~= nil then
	if b*.b_sNext = nil then
	    /* the selected buffer isn't on screen - replace the current one */
	    if CurrentBuffer.b_sNext = BufferPtr then
		/* current is the only window */
		b*.b_sNext := b;
		b*.b_sPrev := b;
	    else
		CurrentBuffer.b_sNext*.b_sPrev := b;
		CurrentBuffer.b_sPrev*.b_sNext := b;
		b*.b_sNext := CurrentBuffer.b_sNext;
		b*.b_sPrev := CurrentBuffer.b_sPrev;
	    fi;
	    b*.b_windowPos := CurrentBuffer.b_windowPos;
	    b*.b_windowHeight := CurrentBuffer.b_windowHeight;
	    ScreenList := b;
	    CurrentBuffer.b_sNext := nil;
	    CurrentBuffer.b_sPrev := nil;
	    BufferPtr* := CurrentBuffer;
	    BufferPtr := b;
	    CurrentBuffer := b*;
	    if CurrentBuffer.b_windowLine >= CurrentBuffer.b_windowHeight then
		/* the window is now smaller than it used to be and the active
		   line is past the end of the window - move the window
		   back to put the active line in the window */
		displayNew();
	    else
		displayWindow();
	    fi;
	    fullStatus();
	else
	    /* just move to the selected window */
	    BufferPtr := b;
	    CurrentBuffer := b*;
	fi;
	newBuffer();
    fi;
corp;

/*
 * goWindow - go to next or prev window.
 */

proc goWindow(bool isNext)void:

    fixLine();
    BufferPtr* := CurrentBuffer;
    BufferPtr :=
	if isNext then
	    CurrentBuffer.b_sNext
	else
	    CurrentBuffer.b_sPrev
	fi;
    CurrentBuffer := BufferPtr*;
    newBuffer();
corp;

/*
 * abandonAll - check all buffers for abandonment.
 */

proc abandonAll()bool:
    register *BUFFERLIST b, first;
    bool changed;

    fixLine();
    BufferPtr* := CurrentBuffer;
    changed := false;
    b := BufferList;
    first := b;
    while
	if b*.b_changeCount ~= 0 then
	    changed := true;
	fi;
	b := b*.b_next;
	b ~= first
    do
    od;
    if changed then
	ask("Abandon all changes? (Y/N) ")
    else
	true
    fi
corp;

/*
 * updateFiles - update all files currently in memory.
 */

proc updateFiles()bool:
    register *BUFFERLIST b, first, this;
    bool problem;

    fixLine();
    problem := false;
    this := BufferPtr;
    this* := CurrentBuffer;
    b := BufferList;
    first := b;
    while
	if b*.b_changeCount ~= 0 then
	    BufferPtr := b;
	    CurrentBuffer := b*;
	    if not updateFile() then
		problem := true;
	    fi;
	fi;
	b := b*.b_next;
	b ~= first
    do
    od;
    BufferPtr := this;
    CurrentBuffer := this*;
    not problem
corp;

/*
 * bufferList - list all active buffers.
 */

proc bufferList()void:
    register *BUFFERLIST b, first;
    register ushort col, len;

    fixLine();
    BufferPtr* := CurrentBuffer;
    b := BufferList;
    first := b;
    putError("");
    col := 0;
    while
	len := CharsLen(b*.b_fileName) +
		if b*.b_changeCount = 0 then 2 else 3 fi;
	if col + len >= NCOLUMNS - 3 then
	    CRT_Move(NLines - 1, NCOLUMNS - 2);
	    CRT_PutChar('+');
	    pretend(CRT_GetChar(), void);
	    CRT_ClearLine(NLines - 1);
	    col := 0;
	fi;
	if b*.b_changeCount ~= 0 then
	    CRT_PutChar('*');
	fi;
	CRT_PutChars(b*.b_fileName);
	col := col + len;
	b := b*.b_next;
	b ~= first
    do
	CRT_PutChars(", ");
    od;
corp;

/***************************************************************************/

/*
 * THE ACTUAL COMMAND HANDLER: A BIG CASE STATEMENT.
 */

proc command0(ushort cmd)void:
    register ushort column, start;

    case cmd
    incase 0:
	/* move cursor left one column */
	cursorLeft();
    incase 1:
	/* move cursor right one column */
	cursorRight();
    incase 2:
	/* move cursor up one line */
	cursorUp();
    incase 3:
	/* move cursor down one line */
	cursorDown();
    incase 4:
	/* move cursor forward to next word */
	unFixLine();
	fixPosition();
	skipWord();
    incase 5:
	/* cursor back one word */
	unFixLine();
	fixPosition();
	backWord();
    incase 6:
	/* move cursor to beginning of current line */
	CurrentBuffer.b_cursorColumn := 0;
	CursorPos := &LineBuffer[0];
    incase 7:
	/* cursor to end of line */
	unFixLine();
	CurrentBuffer.b_cursorColumn := LineLength;
	CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
    incase 8:
	/* move cursor to beginning of next line */
	cursorDown();
	CurrentBuffer.b_cursorColumn := 0;
    incase 9:
	/* move cursor to top of screen */
	fixLine();
	while CurrentBuffer.b_windowLine ~= 0 do
	    cursorUp();
	od;
    incase 10:
	/* move cursor to bottom of screen */
	fixLine();
	while CurrentBuffer.b_cursorLine < CurrentBuffer.b_lineCount and
		CurrentBuffer.b_windowLine ~=
		CurrentBuffer.b_windowHeight - 1 do
	    cursorDown();
	od;
    incase 11:
	/* move backward one page in file */
	pageBackward();
    incase 12:
	/* move forward one page in file */
	pageForward();
    incase 13:
	/* cursor to beginning of file */
	cursorHome();
    incase 14:
	/* go to a specific line in the file */
	gotoLine();
    incase 15:
	/* go to the previously marked position */
	gotoRegion();
    incase 16:
	/* DEL key: delete current character */
	deleteChar();
    incase 17:
	/* backspace key: delete previous character */
	unFixLine();
	fixPosition();
	if CurrentBuffer.b_cursorColumn ~= 0 then
	    backUpOne();
	    deleteChunk(CursorPos, CursorPos + 1, true);
	fi;
    incase 18:
	/* tab key */
	tabKey();
    incase 19:
	/* delete word in forward direction */
	deleteNextWord();
    incase 20:
	/* delete word in backwards direction */
	deletePrevWord();
    incase 21:
	/* delete from current position to the end of the line */
	deleteToEnd();
    incase 22:
	/* insert an arbitrary character */
	insertArb();
    incase 23:
	/* undo changes to current line */
	LineFixed := true;
	CRT_ClearLine(CurrentBuffer.b_windowPos +
		      CurrentBuffer.b_windowLine);
	if CurrentBuffer.b_currentLine ~= nil then
	    displayString(CurrentBuffer.b_currentLine*.l_data);
	fi;
    incase 24:
	/* redraw screen */
	fixLine();
	CRT_Reset();
	displayScreen();
    incase 25:
	/* delete line */
	deleteLine();
    incase 26:
	/* carriage return: insert line or do multi-column return */
	carriageReturn();
    incase 27:
	/* insert line - almost the same as carriage return */
	insertLine();
    incase 28:
	/* join the next line to this line */
	joinLines();
    incase 29:
	/* split the current line into two pieces at this column */
	splitLine();
    incase 30:
	/* increment current indent amount */
	column := CurrentBuffer.b_indentAmount +
		CurrentBuffer.b_indentIncrement;
	if column < CurrentBuffer.b_wrapColumn then
	    CurrentBuffer.b_indentAmount := column;
	    unFixLine();
	    start := 0;
	    while LineBuffer[start] = ' ' do
		start := start + 1;
	    od;
	    if start <= column and CurrentBuffer.b_cursorColumn <= start then
		adjustLine(true);
	    fi;
	fi;
    incase 31:
	/* decrement current indent amount */
	column := CurrentBuffer.b_indentAmount;
	if CurrentBuffer.b_indentAmount >=
		CurrentBuffer.b_indentIncrement then
	    CurrentBuffer.b_indentAmount := CurrentBuffer.b_indentAmount -
		    CurrentBuffer.b_indentIncrement;
	else
	    CurrentBuffer.b_indentAmount := 0;
	fi;
	unFixLine();
	start := 0;
	while LineBuffer[start] = ' ' do
	    start := start + 1;
	od;
	if start >= column and CurrentBuffer.b_cursorColumn <= start then
	    adjustLine(true);
	fi;
    esac;
corp;

/*
 * command - process an editor command.
 */

proc command(ushort cmd)bool:
    *char
	DELOKMESS = "File deleted.",
	DELNOMESS = "No such file.";
    register ushort column;
    bool continue;

    if BufferList = nil and cmd ~= 51 and cmd ~= 52 and cmd ~= 62 and
	cmd ~= 74
    then
	noFile();
	true
    else
	continue := true;
	case cmd
	default:
	    command0(cmd);
	incase 32:
	    /* adjust the current line to the left indent */
	    adjustLine(true);
	incase 33:
	    /* global search and replace */
	    globalChange();
	incase 34:
	    /* query replace mode */
	    queryReplace();
	incase 35:
	    /* start a search in forward direction */
	    find();
	incase 36:
	    /* start a search in revers direction */
	    reverseFind();
	incase 37:
	    /* mark first end for various operations */
	    setPoint();
	incase 38:
	    /* adjust the lines of a region to the current indent amount */
	    adjustRegion(0);
	incase 39:
	    /* shift the selected region left by indent increment */
	    adjustRegion(1);
	incase 40:
	    /* shift the selected region right by indent increment */
	    adjustRegion(2);
	incase 41:
	    /* move a chunk from the file to the text buffer */
	    grabRegion(true);
	incase 42:
	    /* copy the selected region into the text buffer */
	    grabRegion(false);
	incase 43:
	    /* insert from the text buffer into the file */
	    putRegion();
	incase 44:
	    /* the magic format command */
	    formatRegion();
	incase 45:
	    /* print the selected region on the printer */
	    outputRegion(true);
	incase 46:
	    /* write a chunk to a file */
	    outputRegion(false);
	incase 47:
	    /* read a chunk of text from a file */
	    fileRead();
	incase 48:
	    /* delete a disk file */
	    if getFileName("File to delete: ") then
		putError(
		    if DeleteFile(&LineBuffer[0]) ~= 0 then
			DELOKMESS
		    else
			DELNOMESS
		    fi);
	    fi;
	incase 49:
	    /* scroll up */
	    scrollUp();
	incase 50:
	    /* scroll down one line */
	    scrollDown();
	incase 51:
	    /* exit after updating files */
	    if BufferList = nil or ask("Exit (Y/N)? ") then
		fixLine();
		if updateFiles() then
		    continue := false;
		fi;
	    fi;
	incase 52:
	    /* exit without updating files */
	    if BufferList = nil or abandonAll() then
		continue := false;
	    fi;
	incase 53:
	    /* set a numeric parameter */
	    setNumeric();
	incase 54:
	    /* cycle tab expand flag */
	    blankMode();
	incase 55:
	    /* clear all tab stops */
	    for column from 0 upto NCOLUMNS - 1 do
		TabStops[column] := false;
	    od;
	incase 56:
	    /* set a tab stop at this column */
	    TabStops[CurrentBuffer.b_cursorColumn] := true;
	incase 57:
	    /* clear a tab stop at this column */
	    TabStops[CurrentBuffer.b_cursorColumn] := false;
	incase 58:
	incase 59:
	    /* toggle insert mode */
	    CurrentBuffer.b_insertMode := not CurrentBuffer.b_insertMode;
	    insertStatus();
	incase 60:
	    /* toggle word-wrap status */
	    CurrentBuffer.b_wrapMode := not CurrentBuffer.b_wrapMode;
	    wrapStatus();
	incase 61:
	    /* toggle multicolumn mode flag */
	    CurrentBuffer.b_multiColumn := not CurrentBuffer.b_multiColumn;
	    multiStatus();
	incase 62:
	    /* visit a new file */
	    visitFile();
	incase 63:
	    /* delete the current buffer */
	    deleteBuffer();
	incase 64:
	    /* move to a different buffer */
	    otherBuffer();
	incase 65:
	    /* delete the current window */
	    BufferPtr* := CurrentBuffer;
	    deleteWindow();
	    newBuffer();
	incase 66:
	    /* split the current window into two */
	    splitWindow();
	    newBuffer();
	incase 67:
	    /* make the current window the only window */
	    oneWindow();
	incase 68:
	    /* go to the next window */
	    goWindow(true);
	incase 69:
	    /* go to the previous window */
	    goWindow(false);
	incase 70:
	    /* go update all disk files based on current buffers */
	    pretend(updateFiles(), void);
	incase 71:
	    /* list all current buffers */
	    bufferList();
	incase 72:
	    /* move cursor to beginning of previous line */
	    cursorUp();
	    CurrentBuffer.b_cursorColumn := 0;
	incase 73:
	    /* execute file of changes on rest of current file */
	    batchChange();
	incase 74:
	    /* start up a new shell */
	    if not Execute("NewShell con:0/0/662/197/Ded_Shell/CLOSE",0,0) then
		putError("Failed to create shell");
	    fi;
	esac;
	continue
    fi
corp;
