#drinc:util.g
#drinc:crt.g
#ded.g

/*
 * DEDU.DRC - utility routines for the Draco editor.
 */

extern
    update()void,
    noSpace()void,
    tooLong()void,
    unFixLine()void,
    fullStatus()void,
    putError(*char message)void,
    clearError()void;

**LINELIST PrevPtr;
*LINELIST List, Prev;
ulong Count;

/***************************************************************************/

/*
 * SOME SIMPLE, GENERAL UTILITY ROUTINES.
 */

/*
 * freeString - free a dynamically allocated string.
 */

proc freeString(*char p)void:

    Mfree(pretend(p, *byte), CharsLen(p) + 1);
corp;

/*
 * fixPosition - move the cursor to a valid column on the line.
 */

proc fixPosition()void:
    ushort length;

    unFixLine();
    if CurrentBuffer.b_cursorColumn > LineLength then
	CurrentBuffer.b_cursorColumn := LineLength;
	CursorPos := &LineBuffer[CurrentBuffer.b_cursorColumn];
    fi;
corp;

/*
 * initLines - if no lines yet, then initialize the first one.
 */

proc initLines()void:
    *char p;

    /* this routine MUST succeed!!!! */
    if CurrentBuffer.b_lines = nil then
	MerrorSet(false);
	CurrentBuffer.b_lines := new(LINELIST);
	CurrentBuffer.b_currentLine := CurrentBuffer.b_lines;
	CurrentBuffer.b_currentLine*.l_prev := nil;
	CurrentBuffer.b_currentLine*.l_next := nil;
	p := new(char);
	p* := '\e';
	CurrentBuffer.b_currentLine*.l_data := p;
	CurrentBuffer.b_lineCount := 1;
	MerrorSet(true);
    fi;
corp;

/*
 * isBlank - return true if the character under the cursor is a blank.
 */

proc isBlank()bool:

    CursorPos* = ' '
corp;

/*
 * isLetDig - return true if the character under the cursor is a letter
 *	      or a digit.
 */

proc isLetDig()bool:
    register char ch;

    ch := CursorPos*;
    ch >= 'a' and ch <= 'z' or ch >= 'A' and ch <= 'Z' or
	ch >= '0' and ch <= '9' or ch = '"' or ch = '\'' or ch = '_' or
	ch = '^' or ch = '$'
corp;

/*
 * forwardOne - go forward one character (used for word movement, etc.).
 */

proc forwardOne()void:

    CurrentBuffer.b_cursorColumn := CurrentBuffer.b_cursorColumn + 1;
    CursorPos := CursorPos + 1;
corp;

/*
 * backUpOne - back up one character (used for word movement, word wrap).
 */

proc backUpOne()void:

    CurrentBuffer.b_cursorColumn := CurrentBuffer.b_cursorColumn - 1;
    CursorPos := CursorPos - 1;
corp;

/***************************************************************************/

/*
 * A SET OF ROUTINES FOR MANIPULATING LISTS OF LINES.
 */

/*
 * buildStart - start building a list of lines.
 */

proc buildStart()void:

    PrevPtr := &List;
    Prev := nil;
    Count := 0;
corp;

/*
 * buildLine - add a line to the list being built.
 */

proc buildLine(*char textPtr)bool:
    register *LINELIST l;
    *char p;

    l := new(LINELIST);
    if l = nil then
	noSpace();
	false
    else
	p := pretend(Malloc(CharsLen(textPtr) + 1), *char);
	if p = nil then
	    free(l);
	    noSpace();
	    false
	else
	    CharsCopy(p, textPtr);
	    Count := Count + 1;
	    l*.l_prev := Prev;
	    l*.l_data := p;
	    Prev := l;
	    PrevPtr* := l;
	    PrevPtr := &l*.l_next;
	    true
	fi
    fi
corp;

/*
 * buildEnd - end the building of a list of lines. Return pointers to the
 *	      head and tail of the list and the count of the number of lines.
 */

proc buildEnd(**LINELIST pHead, pTail)ulong:

    PrevPtr* := nil;
    pHead* := List;
    pTail* := Prev;
    Count
corp;

/*
 * buildCopy - build and insert a copy of a list of lines.
 */

proc buildCopy(register *LINELIST l; **LINELIST pHead, pTail)ulong:

    buildStart();
    while l ~= nil and buildLine(l*.l_data) do
	l := l*.l_next;
    od;
    buildEnd(pHead, pTail)
corp;

/*
 * clearMemory - release all memory associated with the passed list of lines.
 */

proc clearMemory(register *LINELIST l)void:
    register *LINELIST temp;

    putError("Freeing memory...");
    CRT_Flush();
    while l ~= nil do
	temp := l;
	freeString(l*.l_data);
	l := l*.l_next;
	free(temp);
    od;
    clearError();
corp;

/***************************************************************************/

/*
 * ROUTINES FOR DISPLAYING TEXT ON THE SCREEN.
 */

/*
 * displayChar - display a possibly funny character.
 */

proc displayChar(char ch)void:

    if ch < ' ' or ch >= '\(0x80)' then
	CRT_EnterHighLight();
	if ch >= '\(0x80)' then
	    ch := ch - 128;
	fi;
	CRT_PutChar(if ch < ' ' then ch + 64 else ch fi);
	CRT_ExitHighLight();
    else
	CRT_PutChar(ch);
    fi;
corp;

/*
 * displayChars - display an expanded string on the screen.
 */

proc displayChars(register *char p)void:

    while p* ~= '\e' do
	displayChar(p*);
	p := p + 1;
    od;
corp;

/*
 * displayString - display a compressed string on the screen.
 */

proc displayString(register *char p)void:
    char NOQUOTE = '\(0x00)';
    register char ch, quote;

    quote := NOQUOTE;
    while
	ch := p*;
	p := p + 1;
	ch ~= '\e'
    do
	if ch = quote then
	    quote := NOQUOTE;
	elif ch = '"' and quote = NOQUOTE then
	    quote := ch;
	fi;
	if ch >= '\(0x80)' and quote = NOQUOTE then
	    while ch > '\(0x80)' do
		ch := ch - 1;
		CRT_PutChar(' ');
	    od;
	else
	    displayChar(ch);
	fi;
    od;
corp;

/*
 * displayList - display a list of lines on the screen for a maximum count.
 */

proc displayList(register *LINELIST l; register ushort count)void:

    while l ~= nil and count ~= 0 do
	count := count - 1;
	displayString(l*.l_data);
	CRT_PutChars("\r\n");
	l := l*.l_next;
    od;
corp;

/*
 * displayWindow - display the current full window.
 */

proc displayWindow()void:
    register *LINELIST l;
    register ushort r;

    /* first must find which line will be at the top of the window. */
    l := CurrentBuffer.b_currentLine;
    r := CurrentBuffer.b_windowLine;
    while r ~= 0 do
	r := r - 1;
	l := l*.l_prev;
    od;
    CRT_ClearLines(CurrentBuffer.b_windowPos, CurrentBuffer.b_windowHeight);
    CRT_Move(CurrentBuffer.b_windowPos, 0);
    displayList(l, CurrentBuffer.b_windowHeight);
corp;

/*
 * displayScreen - display the current full screen.
 */

proc displayScreen()void:
    register *BUFFERLIST first, b;

    BufferPtr* := CurrentBuffer;
    b := ScreenList;
    first := b;
    while
	CurrentBuffer := b*;
	displayWindow();
	fullStatus();
	b := b*.b_sNext;
	b ~= first
    do
    od;
    CurrentBuffer := BufferPtr*;
corp;

/*
 * redraw - redraw the window from the current line on down.
 */

proc redraw()void:

    CRT_ClearLines(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,
	CurrentBuffer.b_windowHeight - CurrentBuffer.b_windowLine);
    displayList(CurrentBuffer.b_currentLine,
		CurrentBuffer.b_windowHeight - CurrentBuffer.b_windowLine);
corp;

/*
 * displayBottom - redraw current window from given line to end of window.
 *		   ASSUMPTION: start is beyond the current line.
 */

proc displayBottom(ushort start)void:
    register *LINELIST l;
    register ushort i;
    ushort count;

    count := CurrentBuffer.b_windowHeight - start;
    CRT_ClearLines(CurrentBuffer.b_windowPos + start, count);
    l := CurrentBuffer.b_currentLine;
    i := start - CurrentBuffer.b_windowLine;
    while i ~= 0 and l ~= nil do
	l := l*.l_next;
	i := i - 1;
    od;
    if l ~= nil then
	displayList(l, count);
    fi;
corp;

/*
 * displayNew - display a windowful newly found. Put the current line
 *		one third of the way down the window.
 */

proc displayNew()void:
    register *LINELIST l;
    register ushort i;

    CurrentBuffer.b_windowLine := 0;
    if CurrentBuffer.b_currentLine ~= nil then
	l := CurrentBuffer.b_currentLine;
	i := CurrentBuffer.b_windowHeight / 3;
	while l*.l_prev ~= nil and i ~= 0 do
	    i := i - 1;
	    CurrentBuffer.b_windowLine := CurrentBuffer.b_windowLine + 1;
	    l := l*.l_prev;
	od;
    fi;
    displayWindow();
corp;

/***************************************************************************/

/*
 * NO WHERE ELSE TO PUT THIS GUY!
 */

/*
 * insertLines - insert a group of lines at the current position.
 *		 The lines are inserted BEFORE the current line.
 */

proc insertLines(ulong count; register *LINELIST head, tail)void:
    ushort tailSize;

    if count ~= 0 then
	if CurrentBuffer.b_gotRegion and
		CurrentBuffer.b_regionLine >= CurrentBuffer.b_cursorLine then
	    CurrentBuffer.b_regionLine := CurrentBuffer.b_regionLine + count;
	fi;
	CurrentBuffer.b_changeCount := CurrentBuffer.b_changeCount + 1;
	CurrentBuffer.b_lineCount := CurrentBuffer.b_lineCount + count;
	if CurrentBuffer.b_cursorLine = 1 then
	    CurrentBuffer.b_lines := head;
	else
	    CurrentBuffer.b_currentLine*.l_prev*.l_next := head;
	fi;
	head*.l_prev :=
	    if CurrentBuffer.b_currentLine = nil then
		nil
	    else
		CurrentBuffer.b_currentLine*.l_prev
	    fi;
	tail*.l_next := CurrentBuffer.b_currentLine;
	if CurrentBuffer.b_currentLine ~= nil then
	    CurrentBuffer.b_currentLine*.l_prev := tail;
	fi;
	CurrentBuffer.b_currentLine := head;
	tailSize :=
	    CurrentBuffer.b_windowHeight - CurrentBuffer.b_windowLine;
	if count <= tailSize then
	    CRT_ScrollRegionDown(
		CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,
		tailSize, count);
	else
	    CRT_ClearLines(
		CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,
		tailSize);
	    count := tailSize;
	fi;
	CRT_Move(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine, 0);
	displayList(head, count);
    fi;
corp;

/***************************************************************************/

/*
 * ROUTINES FOR DEALING WITH SMALL CHUNKS OF TEXT IN Linebuffer.
 */

/*
 * deleteChunk - delete a chunk of data between a pair of pointers.
 */

proc deleteChunk(register *char startPos, endPos; bool needRedraw)void:

    if needRedraw then
	CRT_Move(CurrentBuffer.b_windowPos + CurrentBuffer.b_windowLine,
		 startPos - &LineBuffer[0]);
	CRT_DeleteChars(endPos - startPos);
    fi;
    BlockCopy(pretend(startPos, *byte), pretend(endPos, *byte),
	      CharsLen(endPos) + 1);
    LineLength := LineLength - (endPos - startPos);
    LineChanged := true;
corp;

/*
 * addText - add some text into LineBuffer at CursorPos. Display it.
 */

proc addText(register *char p; bool needRedraw)void:

    if needRedraw then
	displayChars(p);
    fi;
    while p* ~= '\e' do
	CursorPos* := p*;
	p := p + 1;
	forwardOne();
    od;
    LineChanged := true;
corp;

/*
 * insertChunk - insert some stuff at the current position.
 */

proc insertChunk(*char newText; bool needRedraw)void:
    ushort extraLen, newLen;

    unFixLine();
    fixPosition();
    extraLen := CharsLen(newText);
    newLen := extraLen + LineLength;
    if newLen > NCOLUMNS - 1 then
	tooLong();
    else
	BlockCopyB(pretend(&LineBuffer[newLen], *byte),
		   pretend(&LineBuffer[LineLength], *byte),
		   LineLength - CurrentBuffer.b_cursorColumn + 1);
	if needRedraw then
	    CRT_InsertChars(extraLen);
	    update();
	fi;
	addText(newText, needRedraw);
	LineLength := newLen;
    fi;
corp;

/*
 * insertBlanks - insert a string of N blanks.
 */

proc insertBlanks(ushort n; bool wantRedraw)void:
    register ushort newLen, i;

    unFixLine();
    fixPosition();
    newLen := LineLength + n;
    if newLen > NCOLUMNS - 1 then
	tooLong();
    else
	BlockCopyB(pretend(&LineBuffer[newLen], *byte),
		   pretend(&LineBuffer[LineLength], *byte),
		   LineLength - CurrentBuffer.b_cursorColumn + 1);
	i := n;
	while i ~= 0 do
	    i := i - 1;
	    CursorPos* := ' ';
	    forwardOne();
	od;
	LineChanged := true;
	LineLength := newLen;
	if wantRedraw then
	    CRT_InsertChars(n);
	    update();
	fi;
    fi;
corp;
